# Copyright 2017,2018 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Handler for the root of the sdk API."""

import json
import six
import threading
import webob.exc

from zvmconnector import connector
from zvmsdk import config
from zvmsdk import log
from zvmsdk import returncode
from zvmsdk.sdkwsgi.handlers import tokens
from zvmsdk.sdkwsgi.schemas import guest
from zvmsdk.sdkwsgi import util
from zvmsdk.sdkwsgi import validation
from zvmsdk import utils


_VMACTION = None
_VMHANDLER = None
CONF = config.CONF
LOG = log.LOG
CONF = config.CONF


class VMHandler(object):
    def __init__(self):
        self.client = connector.ZVMConnector(connection_type='socket',
                                             ip_addr=CONF.sdkserver.bind_addr,
                                             port=CONF.sdkserver.bind_port)

    @validation.schema(guest.create)
    def create(self, body):
        guest = body['guest']

        userid = guest['userid']
        vcpus = guest['vcpus']
        memory = guest['memory']

        kwargs_list = {}
        guest_keys = guest.keys()
        if 'disk_list' in guest_keys:
            kwargs_list['disk_list'] = guest['disk_list']
        if 'user_profile' in guest_keys:
            kwargs_list['user_profile'] = guest['user_profile']
        if 'max_cpu' in guest_keys:
            kwargs_list['max_cpu'] = guest['max_cpu']
        if 'max_mem' in guest_keys:
            kwargs_list['max_mem'] = guest['max_mem']

        info = self.client.send_request('guest_create', userid, vcpus,
                                        memory, **kwargs_list)

        return info

    def list(self):
        # list all guest on the given host
        info = self.client.send_request('guest_list')
        return info

    @validation.query_schema(guest.userid_list_query)
    def get_info(self, req, userid):
        info = self.client.send_request('guest_get_info', userid)
        return info

    @validation.query_schema(guest.userid_list_query)
    def get_definition_info(self, req, userid):
        info = self.client.send_request('guest_get_definition_info', userid)
        return info

    @validation.query_schema(guest.userid_list_query)
    def get_power_state(self, req, userid):
        info = self.client.send_request('guest_get_power_state', userid)
        return info

    def delete(self, userid):
        info = self.client.send_request('guest_delete', userid)
        return info

    def delete_nic(self, userid, vdev, body):
        active = body.get('active', False)
        active = util.bool_from_string(active, strict=True)

        info = self.client.send_request('guest_delete_nic', userid, vdev,
                                        active=active)
        return info

    @validation.query_schema(guest.userid_list_array_query)
    def inspect_stats(self, req, userid_list):
        info = self.client.send_request('guest_inspect_stats',
                                        userid_list)
        return info

    @validation.query_schema(guest.userid_list_array_query)
    def inspect_vnics(self, req, userid_list):
        info = self.client.send_request('guest_inspect_vnics',
                                        userid_list)
        return info

    # @validation.query_schema(guest.nic_DB_info)
    # FIXME: the above validation will fail with "'dict' object has no
    # attribute 'dict_of_lists'"
    def get_nic_DB_info(self, req, userid=None, nic_id=None, vswitch=None):
        info = self.client.send_request('guests_get_nic_info', userid=userid,
                                        nic_id=nic_id, vswitch=vswitch)
        return info

    @validation.schema(guest.create_nic)
    def create_nic(self, userid, body=None):
        nic = body['nic']

        vdev = nic.get('vdev', None)
        nic_id = nic.get('nic_id', None)
        mac_addr = nic.get('mac_addr', None)
        active = nic.get('active', False)
        active = util.bool_from_string(active, strict=True)

        info = self.client.send_request('guest_create_nic', userid,
                                        vdev=vdev, nic_id=nic_id,
                                        mac_addr=mac_addr,
                                        active=active)
        return info

    @validation.schema(guest.create_network_interface)
    def create_network_interface(self, userid, body=None):
        interface = body['interface']
        version = interface['os_version']
        networks = interface.get('guest_networks', None)
        active = interface.get('active', False)
        active = util.bool_from_string(active, strict=True)
        info = self.client.send_request('guest_create_network_interface',
                                        userid, os_version=version,
                                        guest_networks=networks,
                                        active=active)
        return info

    @validation.schema(guest.delete_network_interface)
    def delete_network_interface(self, userid, body=None):
        interface = body['interface']
        version = interface['os_version']
        vdev = interface['vdev']
        active = interface.get('active', False)
        active = util.bool_from_string(active, strict=True)
        info = self.client.send_request('guest_delete_network_interface',
                                        userid, version, vdev,
                                        active=active)
        return info

    @validation.schema(guest.create_disks)
    def create_disks(self, userid, body=None):
        disk_info = body['disk_info']
        disk_list = disk_info.get('disk_list', None)
        info = self.client.send_request('guest_create_disks', userid,
                                        disk_list)
        return info

    @validation.schema(guest.config_minidisks)
    def config_minidisks(self, userid, body=None):
        disk_info = body['disk_info']
        disk_list = disk_info.get('disk_list', None)
        info = self.client.send_request('guest_config_minidisks', userid,
                                        disk_list)
        return info

    @validation.schema(guest.delete_disks)
    def delete_disks(self, userid, body=None):
        vdev_info = body['vdev_info']
        vdev_list = vdev_info.get('vdev_list', None)

        info = self.client.send_request('guest_delete_disks', userid,
                                        vdev_list)

        return info

    @validation.schema(guest.nic_couple_uncouple)
    def nic_couple_uncouple(self, userid, vdev, body):
        info = body['info']

        active = info.get('active', False)
        active = util.bool_from_string(active, strict=True)

        couple = util.bool_from_string(info['couple'], strict=True)

        if couple:
            info = self.client.send_request('guest_nic_couple_to_vswitch',
                                            userid, vdev, info['vswitch'],
                                            active=active)
        else:
            info = self.client.send_request('guest_nic_uncouple_from_vswitch',
                                            userid, vdev,
                                            active=active)
        return info


class VMAction(object):

    def __init__(self):
        self.client = connector.ZVMConnector(connection_type='socket',
                                             ip_addr=CONF.sdkserver.bind_addr,
                                             port=CONF.sdkserver.bind_port)
        self.dd_semaphore = threading.BoundedSemaphore(
            value=CONF.wsgi.max_concurrent_deploy_capture)

    def start(self, userid, body):
        info = self.client.send_request('guest_start', userid)

        return info

    @validation.schema(guest.stop)
    def stop(self, userid, body):
        timeout = body.get('timeout', None)
        poll_interval = body.get('poll_interval', None)

        info = self.client.send_request('guest_stop', userid,
                                        timeout=timeout,
                                        poll_interval=poll_interval)

        return info

    @validation.schema(guest.softstop)
    def softstop(self, userid, body):
        timeout = body.get('timeout', None)
        poll_interval = body.get('poll_interval', None)

        info = self.client.send_request('guest_softstop', userid,
                                        timeout=timeout,
                                        poll_interval=poll_interval)

        return info

    def pause(self, userid, body):
        info = self.client.send_request('guest_pause', userid)

        return info

    def unpause(self, userid, body):
        info = self.client.send_request('guest_unpause', userid)

        return info

    def reboot(self, userid, body):
        info = self.client.send_request('guest_reboot', userid)

        return info

    def reset(self, userid, body):
        info = self.client.send_request('guest_reset', userid)

        return info

    def get_console_output(self, userid, body):
        info = self.client.send_request('guest_get_console_output',
                                        userid)

        return info

    @validation.schema(guest.register_vm)
    def register_vm(self, userid, body):
        meta = body['meta']
        net_set = body['net_set']
        info = self.client.send_request('guest_register',
                                userid, meta, net_set)
        return info

    @validation.schema(guest.live_migrate_vm)
    def live_migrate_vm(self, userid, body):
        # dest_zcc_userid default as ''
        dest_zcc_userid = body.get('dest_zcc_userid', '')
        destination = body['destination']
        operation = body.get('operation', {})
        parms = body['parms']

        info = self.client.send_request('guest_live_migrate',
                                userid, dest_zcc_userid, destination,
                                parms, operation)
        return info

    @validation.schema(guest.resize_cpus)
    def resize_cpus(self, userid, body):
        cpu_cnt = body['cpu_cnt']
        info = self.client.send_request('guest_resize_cpus',
                                        userid, cpu_cnt)

        return info

    @validation.schema(guest.resize_cpus)
    def live_resize_cpus(self, userid, body):
        cpu_cnt = body['cpu_cnt']
        info = self.client.send_request('guest_live_resize_cpus',
                                        userid, cpu_cnt)

        return info

    @validation.schema(guest.resize_mem)
    def resize_mem(self, userid, body):
        size = body['size']
        info = self.client.send_request('guest_resize_mem',
                                        userid, size)

        return info

    @validation.schema(guest.resize_mem)
    def live_resize_mem(self, userid, body):
        size = body['size']
        info = self.client.send_request('guest_live_resize_mem',
                                        userid, size)

        return info

    @validation.schema(guest.deploy)
    def deploy(self, userid, body):
        image_name = body['image']

        transportfiles = body.get('transportfiles', None)
        remotehost = body.get('remotehost', None)
        vdev = body.get('vdev', None)

        request_info = ("action: 'deploy', userid: %(userid)s,"
                        "transportfiles: %(trans)s, remotehost: %(remote)s,"
                        "vdev: %(vdev)s" %
                        {'userid': userid, 'trans': transportfiles,
                         'remote': remotehost, 'vdev': vdev
                         })

        info = None
        dd_allowed = self.dd_semaphore.acquire(blocking=False)
        if not dd_allowed:
            error_def = returncode.errors['serviceUnavail']
            info = error_def[0]
            err_msg = error_def[1][1] % {'req': request_info}
            info.update({'rs': 1,
                         'errmsg': err_msg,
                         'output': ''})
            LOG.error(err_msg)
            return info

        try:
            LOG.debug("WSGI sending deploy requests. %s" % request_info)
            info = self.client.send_request('guest_deploy', userid,
                                            image_name,
                                            transportfiles=transportfiles,
                                            remotehost=remotehost,
                                            vdev=vdev)
        finally:
            try:
                self.dd_semaphore.release()
                LOG.debug("WSGI deploy request finished, %s."
                          "Resource released." % request_info)
            except Exception as err:
                err_msg = ("Failed to release deploy resource in WSGI."
                           "Error: %s, request info: %s" %
                           (six.text_type(err), request_info))
                LOG.error(err_msg)

        return info

    @validation.schema(guest.capture)
    def capture(self, userid, body):
        image_name = body['image']

        capture_type = body.get('capture_type', 'rootonly')
        compress_level = body.get('compress_level', 6)

        request_info = ("action: 'capture', userid: %(userid)s,"
                        "image name: %(image)s, capture type: %(cap)s,"
                        "compress level: %(level)s" %
                        {'userid': userid, 'image': image_name,
                         'cap': capture_type, 'level': compress_level
                         })
        info = None
        capture_allowed = self.dd_semaphore.acquire(blocking=False)
        if not capture_allowed:
            error_def = returncode.errors['serviceUnavail']
            info = error_def[0]
            err_msg = error_def[1][1] % {'req': request_info}
            info.update({'rs': 1,
                         'errmsg': err_msg,
                         'output': ''})
            LOG.error(err_msg)
            return info

        try:
            LOG.debug("WSGI sending capture requests. %s" % request_info)
            info = self.client.send_request('guest_capture', userid,
                                            image_name,
                                            capture_type=capture_type,
                                            compress_level=compress_level)
        finally:
            try:
                self.dd_semaphore.release()
                LOG.debug("WSGI capture request finished, %s."
                          "Resource released." % request_info)
            except Exception as err:
                err_msg = ("Failed to release capture resource in WSGI."
                           "Error: %s, request info: %s" %
                           (six.text_type(err), request_info))
                LOG.error(err_msg)

        return info


def get_action():
    global _VMACTION
    if _VMACTION is None:
        _VMACTION = VMAction()
    return _VMACTION


def get_handler():
    global _VMHANDLER
    if _VMHANDLER is None:
        _VMHANDLER = VMHandler()
    return _VMHANDLER


@util.SdkWsgify
@tokens.validate
def guest_get_info(req):

    def _guest_get_info(req, userid):
        action = get_handler()
        return action.get_info(req, userid)

    userid = util.wsgi_path_item(req.environ, 'userid')
    info = _guest_get_info(req, userid)

    info_json = json.dumps(info)
    req.response.status = util.get_http_code_from_sdk_return(info,
        additional_handler=util.handle_not_found)
    req.response.body = utils.to_utf8(info_json)
    req.response.content_type = 'application/json'
    return req.response


@util.SdkWsgify
@tokens.validate
def guest_get(req):

    def _guest_get(req, userid):
        action = get_handler()
        return action.get_definition_info(req, userid)

    userid = util.wsgi_path_item(req.environ, 'userid')
    info = _guest_get(req, userid)

    # info we got looks like:
    # {'user_direct': [u'USER RESTT305 PASSW0RD 1024m 1024m G',
    #                  u'INCLUDE OSDFLT']}
    info_json = json.dumps(info)
    req.response.status = util.get_http_code_from_sdk_return(info,
        additional_handler=util.handle_not_found)
    req.response.body = utils.to_utf8(info_json)
    req.response.content_type = 'application/json'
    return req.response


@util.SdkWsgify
@tokens.validate
def guest_get_power_state(req):

    def _guest_get_power_state(req, userid):
        action = get_handler()
        return action.get_power_state(req, userid)

    userid = util.wsgi_path_item(req.environ, 'userid')
    info = _guest_get_power_state(req, userid)

    info_json = json.dumps(info)
    req.response.status = util.get_http_code_from_sdk_return(info,
        additional_handler=util.handle_not_found)
    req.response.body = utils.to_utf8(info_json)
    req.response.content_type = 'application/json'
    return req.response


@util.SdkWsgify
@tokens.validate
def guest_create(req):

    def _guest_create(req):
        action = get_handler()
        body = util.extract_json(req.body)

        return action.create(body=body)

    info = _guest_create(req)

    info_json = json.dumps(info)
    req.response.body = utils.to_utf8(info_json)
    req.response.status = util.get_http_code_from_sdk_return(info,
        additional_handler=util.handle_already_exists)
    req.response.content_type = 'application/json'
    return req.response


@util.SdkWsgify
@tokens.validate
def guest_list(req):
    def _guest_list():
        action = get_handler()
        return action.list()

    info = _guest_list()
    info_json = json.dumps(info)
    req.response.body = utils.to_utf8(info_json)
    req.response.content_type = 'application/json'
    req.response.status = 200
    return req.response


@util.SdkWsgify
@tokens.validate
def guest_action(req):

    def _guest_action(userid, req):
        action = get_action()
        body = util.extract_json(req.body)
        if len(body) == 0 or 'action' not in body:
            msg = 'action not exist or is empty'
            LOG.info(msg)
            raise webob.exc.HTTPBadRequest(explanation=msg)

        method = body['action']
        func = getattr(action, method, None)
        if func:
            body.pop('action')
            return func(userid, body=body)
        else:
            msg = 'action %s is invalid' % method
            raise webob.exc.HTTPBadRequest(msg)

    userid = util.wsgi_path_item(req.environ, 'userid')
    info = _guest_action(userid, req)

    info_json = json.dumps(info)
    req.response.body = utils.to_utf8(info_json)
    req.response.content_type = 'application/json'
    req.response.status = util.get_http_code_from_sdk_return(info,
        additional_handler=util.handle_not_found)
    return req.response


@util.SdkWsgify
@tokens.validate
def guest_delete(req):

    def _guest_delete(userid):
        action = get_handler()
        return action.delete(userid)

    userid = util.wsgi_path_item(req.environ, 'userid')
    info = _guest_delete(userid)

    info_json = json.dumps(info)
    req.response.body = utils.to_utf8(info_json)
    req.response.status = util.get_http_code_from_sdk_return(info, default=200)
    req.response.content_type = 'application/json'
    return req.response


@util.SdkWsgify
@tokens.validate
def guest_delete_nic(req):
    def _guest_delete_nic(userid, vdev, req):
        action = get_handler()
        body = util.extract_json(req.body)

        return action.delete_nic(userid, vdev, body)

    userid = util.wsgi_path_item(req.environ, 'userid')
    vdev = util.wsgi_path_item(req.environ, 'vdev')

    info = _guest_delete_nic(userid, vdev, req)

    info_json = json.dumps(info)
    req.response.body = utils.to_utf8(info_json)
    req.response.status = util.get_http_code_from_sdk_return(info, default=200)
    req.response.content_type = 'application/json'
    return req.response


@util.SdkWsgify
@tokens.validate
def guest_create_nic(req):

    def _guest_create_nic(userid, req):
        action = get_handler()
        body = util.extract_json(req.body)

        return action.create_nic(userid, body=body)

    userid = util.wsgi_path_item(req.environ, 'userid')
    info = _guest_create_nic(userid, req)

    info_json = json.dumps(info)
    req.response.body = utils.to_utf8(info_json)
    req.response.status = util.get_http_code_from_sdk_return(info)
    req.response.content_type = 'application/json'
    return req.response


@util.SdkWsgify
@tokens.validate
def guest_couple_uncouple_nic(req):

    def _guest_couple_uncouple_nic(userid, vdev, req):
        action = get_handler()
        body = util.extract_json(req.body)

        return action.nic_couple_uncouple(userid, vdev, body=body)

    userid = util.wsgi_path_item(req.environ, 'userid')
    vdev = util.wsgi_path_item(req.environ, 'vdev')

    info = _guest_couple_uncouple_nic(userid, vdev, req)

    info_json = json.dumps(info)
    req.response.body = utils.to_utf8(info_json)
    req.response.status = util.get_http_code_from_sdk_return(info)
    req.response.content_type = 'application/json'
    return req.response


@util.SdkWsgify
@tokens.validate
def guest_create_network_interface(req):

    def _guest_create_network_interface(userid, req):
        action = get_handler()
        body = util.extract_json(req.body)

        return action.create_network_interface(userid, body=body)

    userid = util.wsgi_path_item(req.environ, 'userid')
    info = _guest_create_network_interface(userid, req)

    info_json = json.dumps(info)
    req.response.body = utils.to_utf8(info_json)
    req.response.status = util.get_http_code_from_sdk_return(info)
    req.response.content_type = 'application/json'
    return req.response


@util.SdkWsgify
@tokens.validate
def guest_delete_network_interface(req):

    def _guest_delete_network_interface(userid, req):
        action = get_handler()
        body = util.extract_json(req.body)

        return action.delete_network_interface(userid, body=body)

    userid = util.wsgi_path_item(req.environ, 'userid')
    info = _guest_delete_network_interface(userid, req)

    info_json = json.dumps(info)
    req.response.body = utils.to_utf8(info_json)
    req.response.status = util.get_http_code_from_sdk_return(info)
    req.response.content_type = 'application/json'
    return req.response


def _get_userid_list(req):
    userids = []
    if 'userid' in req.GET.keys():
        userid = req.GET.get('userid')
        userid = userid.strip(' ,')
        userid = userid.replace(' ', '')
        userids.extend(userid.split(','))

    return userids


@util.SdkWsgify
@tokens.validate
def guest_get_stats(req):

    userid_list = _get_userid_list(req)

    def _guest_get_stats(req, userid_list):
        action = get_handler()
        return action.inspect_stats(req, userid_list)

    info = _guest_get_stats(req, userid_list)

    info_json = json.dumps(info)
    req.response.status = util.get_http_code_from_sdk_return(info,
        additional_handler=util.handle_not_found)
    req.response.body = utils.to_utf8(info_json)
    req.response.content_type = 'application/json'
    return req.response


@util.SdkWsgify
@tokens.validate
def guest_get_interface_stats(req):

    userid_list = _get_userid_list(req)

    def _guest_get_interface_stats(req, userid_list):
        action = get_handler()
        return action.inspect_vnics(req, userid_list)

    info = _guest_get_interface_stats(req, userid_list)

    info_json = json.dumps(info)
    req.response.status = util.get_http_code_from_sdk_return(info,
        additional_handler=util.handle_not_found)
    req.response.body = utils.to_utf8(info_json)
    req.response.content_type = 'application/json'
    return req.response


@util.SdkWsgify
@tokens.validate
def guests_get_nic_info(req):

    def _guests_get_nic_DB_info(req, userid=None, nic_id=None, vswitch=None):
        action = get_handler()
        return action.get_nic_DB_info(req, userid=userid, nic_id=nic_id,
                                      vswitch=vswitch)
    userid = req.GET.get('userid', None)
    nic_id = req.GET.get('nic_id', None)
    vswitch = req.GET.get('vswitch', None)

    info = _guests_get_nic_DB_info(req, userid=userid, nic_id=nic_id,
                                   vswitch=vswitch)

    info_json = json.dumps(info)
    req.response.status = util.get_http_code_from_sdk_return(info,
        additional_handler=util.handle_not_found)
    req.response.body = utils.to_utf8(info_json)
    req.response.content_type = 'application/json'
    return req.response


@util.SdkWsgify
@tokens.validate
def guest_config_disks(req):

    def _guest_config_minidisks(userid, req):
        action = get_handler()
        body = util.extract_json(req.body)
        return action.config_minidisks(userid, body=body)

    userid = util.wsgi_path_item(req.environ, 'userid')

    info = _guest_config_minidisks(userid, req)

    info_json = json.dumps(info)
    req.response.status = util.get_http_code_from_sdk_return(info)
    req.response.body = utils.to_utf8(info_json)
    req.response.content_type = 'application/json'
    return req.response


@util.SdkWsgify
@tokens.validate
def guest_create_disks(req):

    def _guest_create_disks(userid, req):
        action = get_handler()
        body = util.extract_json(req.body)
        return action.create_disks(userid, body=body)

    userid = util.wsgi_path_item(req.environ, 'userid')

    info = _guest_create_disks(userid, req)

    info_json = json.dumps(info)
    req.response.status = util.get_http_code_from_sdk_return(info)
    req.response.body = utils.to_utf8(info_json)
    req.response.content_type = 'application/json'
    return req.response


@util.SdkWsgify
@tokens.validate
def guest_delete_disks(req):

    def _guest_delete_disks(userid, req):
        action = get_handler()
        body = util.extract_json(req.body)
        return action.delete_disks(userid, body=body)

    userid = util.wsgi_path_item(req.environ, 'userid')

    info = _guest_delete_disks(userid, req)

    info_json = json.dumps(info)
    req.response.status = util.get_http_code_from_sdk_return(info, default=200)
    req.response.body = utils.to_utf8(info_json)
    req.response.content_type = 'application/json'
    return req.response



# Copyright 2017,2018 IBM Corp.
# Copyright 2013 NEC Corporation.
# Copyright 2011 OpenStack Foundation.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import errno
import functools
import netaddr
import os
import pwd
import re
import shlex
import shutil
import six
import subprocess
import sys
import tempfile
import time
import traceback

from zvmsdk import config
from zvmsdk import constants
from zvmsdk import exception
from zvmsdk import log


CONF = config.CONF
LOG = log.LOG


def execute(cmd):
    """ execute command, return rc and output string.
    The cmd argument can be a string or a list composed of
    the command name and each of its argument.
    eg, ['/usr/bin/cp', '-r', 'src', 'dst'] """

    # Parse cmd string to a list
    if not isinstance(cmd, list):
        cmd = shlex.split(cmd)
    # Execute command
    rc = 0
    output = ""
    try:
        output = subprocess.check_output(cmd, close_fds=True,
                                         stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as err:
        rc = err.returncode
        output = err.output
    except Exception as err:
        err_msg = ('Command "%s" Error: %s' % (' '.join(cmd), str(err)))
        raise exception.SDKInternalError(msg=err_msg)

    output = bytes.decode(output)
    return (rc, output)


def get_host():
    return ''.join([pwd.getpwuid(os.geteuid()).pw_name, '@',
                    CONF.network.my_ip])


def looping_call(f, sleep=5, inc_sleep=0, max_sleep=60, timeout=600,
                 exceptions=(), *args, **kwargs):
    """Helper function that to run looping call with fixed/dynamical interval.

    :param f:          the looping call function or method.
    :param sleep:      initial interval of the looping calls.
    :param inc_sleep:  sleep time increment, default as 0.
    :param max_sleep:  max sleep time.
    :param timeout:    looping call timeout in seconds, 0 means no timeout.
    :param exceptions: exceptions that trigger re-try.

    """
    time_start = time.time()
    expiration = time_start + timeout

    retry = True
    while retry:
        expired = timeout and (time.time() > expiration)
        LOG.debug(
                "timeout is %(timeout)s, expiration is %(expiration)s, \
                        time_start is %(time_start)s" %
                        {"timeout": timeout, "expiration": expiration,
                            "time_start": time_start})

        try:
            f(*args, **kwargs)
        except exceptions:
            retry = not expired
            if retry:
                LOG.debug("Will re-try %(fname)s in %(itv)d seconds" %
                          {'fname': f.__name__, 'itv': sleep})
                time.sleep(sleep)
                sleep = min(sleep + inc_sleep, max_sleep)
            else:
                LOG.debug("Looping call %s timeout" % f.__name__)
            continue
        retry = False


def convert_to_mb(s):
    """Convert memory size from GB to MB."""
    s = s.upper()
    try:
        if s.endswith('G'):
            return float(s[:-1].strip()) * 1024
        elif s.endswith('T'):
            return float(s[:-1].strip()) * 1024 * 1024
        else:
            return float(s[:-1].strip())
    except (IndexError, ValueError, KeyError, TypeError):
        errmsg = ("Invalid memory format: %s") % s
        raise exception.SDKInternalError(msg=errmsg)


class PathUtils(object):

    def clean_temp_folder(self, tmp_folder):
        if os.path.isdir(tmp_folder):
            LOG.debug('Removing existing folder %s ', tmp_folder)
            shutil.rmtree(tmp_folder)

    def _get_guest_path(self):
        return os.path.join(constants.SDK_DATA_PATH, 'guests')

    def mkdir_if_not_exist(self, folder):
        if not os.path.exists(folder):
            LOG.debug("Creating the guest path %s", folder)
            os.makedirs(folder)

    # This is for persistent info for guests
    # by default it's /var/lib/zvmsdk/guests/xxxx
    def remove_guest_path(self, userid):
        guest_folder = os.path.join(self._get_guest_path(), userid)
        try:
            shutil.rmtree(guest_folder)
        except Exception:
            # Ignore any exception for delete temp folder
            pass

    def get_guest_temp_path(self, userid):
        tmp_inst_dir = tempfile.mkdtemp(prefix=userid,
                                        dir='/tmp')
        return tmp_inst_dir

    def get_guest_path(self, userid):
        guest_folder = os.path.join(self._get_guest_path(), userid)
        self.mkdir_if_not_exist(guest_folder)
        return guest_folder

    def get_console_log_path(self, userid):
        return os.path.join(self.get_guest_path(userid), "console.log")

    def create_import_image_repository(self, image_osdistro, type,
                                       image_name):
        zvmsdk_image_import_repo = os.path.join(
                                    CONF.image.sdk_image_repository,
                                    type,
                                    image_osdistro,
                                    image_name)

        if not os.path.exists(zvmsdk_image_import_repo):
            LOG.debug('Creating image repository %s for image import',
                      zvmsdk_image_import_repo)
            os.makedirs(zvmsdk_image_import_repo)
        return zvmsdk_image_import_repo

    def create_file_repository(self, file_type):
        zvmsdk_file_repo = os.path.join(CONF.file.file_repository,
                                        file_type)

        if not os.path.exists(zvmsdk_file_repo):
            LOG.debug('Creating file repository %s for file transfer',
                      zvmsdk_file_repo)
            os.makedirs(zvmsdk_file_repo)
        return zvmsdk_file_repo


def to_utf8(text):
    if isinstance(text, bytes):
        return text
    elif isinstance(text, six.text_type):
        return text.encode()
    else:
        raise TypeError("bytes or Unicode expected, got %s"
                        % type(text).__name__)


def valid_userid(userid):
    if not isinstance(userid, six.string_types):
        return False
    if ((userid == '') or
        (userid.find(' ') != -1)):
        return False
    if len(userid) > 8:
        return False
    return True


def valid_mac_addr(addr):
    ''' Validates a mac address'''
    if not isinstance(addr, six.string_types):
        return False
    valid = re.compile(r'''
                      (^([0-9A-F]{2}[:]){5}([0-9A-F]{2})$)
                      ''',
                      re.VERBOSE | re.IGNORECASE)
    return valid.match(addr) is not None


def valid_cidr(cidr):
    if not isinstance(cidr, six.string_types):
        return False
    try:
        netaddr.IPNetwork(cidr)
    except netaddr.AddrFormatError:
        return False
    if '/' not in cidr:
        return False
    if re.search('\s', cidr):
        return False
    return True


def last_bytes(file_like_object, num):
    try:
        file_like_object.seek(-num, os.SEEK_END)
    except IOError as e:
        # seek() fails with EINVAL when trying to go before the start of the
        # file. It means that num is larger than the file size, so just
        # go to the start.
        if e.errno == errno.EINVAL:
            file_like_object.seek(0, os.SEEK_SET)
        else:
            raise

    remaining = file_like_object.tell()
    return (file_like_object.read(), remaining)


def check_input_types(*types, **validkeys):
    """This is a function decorator to check all input parameters given to
    decorated function are in expected types.

    The checks can be skipped by specify skip_input_checks=True in decorated
    function.

    :param tuple types: expected types of input parameters to the decorated
                        function
    :param validkeys: valid keywords(str) in a list.
                      e.g. validkeys=['key1', 'key2']
    """
    def decorator(function):
        @functools.wraps(function)
        def wrap_func(*args, **kwargs):
            if args[0]._skip_input_check:
                # skip input check
                return function(*args, **kwargs)
            # drop class object self
            inputs = args[1:]
            if (len(inputs) > len(types)):
                msg = ("Too many parameters provided: %(specified)d specified,"
                       "%(expected)d expected." %
                       {'specified': len(inputs), 'expected': len(types)})
                LOG.info(msg)
                raise exception.SDKInvalidInputNumber(function.__name__,
                                                      len(types), len(inputs))

            argtypes = tuple(map(type, inputs))
            match_types = types[0:len(argtypes)]

            invalid_type = False
            invalid_userid_idx = -1
            for idx in range(len(argtypes)):
                _mtypes = match_types[idx]
                if not isinstance(_mtypes, tuple):
                    _mtypes = (_mtypes,)

                argtype = argtypes[idx]
                if constants._TUSERID in _mtypes:
                    userid_type = True
                    for _tmtype in _mtypes:
                        if ((argtype == _tmtype) and
                            (_tmtype != constants._TUSERID)):
                            userid_type = False
                    if (userid_type and
                        (not valid_userid(inputs[idx]))):
                        invalid_userid_idx = idx
                        break
                elif argtype not in _mtypes:
                    invalid_type = True
                    break

            if invalid_userid_idx != -1:
                msg = ("Invalid string value found at the #%d parameter, "
                       "length should be less or equal to 8 and should not be "
                       "null or contain spaces." % (invalid_userid_idx + 1))
                LOG.info(msg)
                raise exception.SDKInvalidInputFormat(msg=msg)

            if invalid_type:
                msg = ("Invalid input types: %(argtypes)s; "
                       "Expected types: %(types)s" %
                       {'argtypes': str(argtypes), 'types': str(types)})
                LOG.info(msg)
                raise exception.SDKInvalidInputTypes(function.__name__,
                                                     str(types), str(argtypes))

            valid_keys = validkeys.get('valid_keys')
            if valid_keys:
                for k in kwargs.keys():
                    if k not in valid_keys:
                        msg = ("Invalid keyword: %(key)s; "
                               "Expected keywords are: %(keys)s" %
                               {'key': k, 'keys': str(valid_keys)})
                        LOG.info(msg)
                        raise exception.SDKInvalidInputFormat(msg=msg)
            return function(*args, **kwargs)
        return wrap_func
    return decorator


def import_class(import_str):
    """Returns a class from a string including module and class."""
    mod_str, _sep, class_str = import_str.rpartition('.')
    __import__(mod_str)
    try:
        return getattr(sys.modules[mod_str], class_str)
    except AttributeError:
        raise ImportError('Class %s cannot be found (%s)' %
                          (class_str,
                           traceback.format_exception(*sys.exc_info())))


def import_object(import_str, *args, **kwargs):
    """Import a class and return an instance of it."""
    return import_class(import_str)(*args, **kwargs)


@contextlib.contextmanager
def expect_invalid_resp_data(data=''):
    """Catch exceptions when using zvm client response data."""
    try:
        yield
    except (ValueError, TypeError, IndexError, AttributeError,
            KeyError) as err:
        msg = ('Invalid smut response data: %s. Error: %s' %
               (data, six.text_type(err)))
        LOG.error(msg)
        raise exception.SDKInternalError(msg=msg)


def wrap_invalid_resp_data_error(function):
    """Catch exceptions when using zvm client response data."""

    @functools.wraps(function)
    def decorated_function(*arg, **kwargs):
        try:
            return function(*arg, **kwargs)
        except (ValueError, TypeError, IndexError, AttributeError,
                KeyError) as err:
            msg = ('Invalid smut response data. Error: %s' %
                   six.text_type(err))
            LOG.error(msg)
            raise exception.SDKInternalError(msg=msg)

    return decorated_function


@contextlib.contextmanager
def expect_and_reraise_internal_error(modID='SDK'):
    """Catch all kinds of zvm client request failure and reraise.

    modID: the moduleID that the internal error happens in.
    """
    try:
        yield
    except exception.SDKInternalError as err:
        msg = err.format_message()
        raise exception.SDKInternalError(msg, modID=modID)


@contextlib.contextmanager
def log_and_reraise_sdkbase_error(action):
    """Catch SDK base exception and print error log before reraise exception.

    msg: the error message to be logged.
    """
    try:
        yield
    except exception.SDKBaseException:
        msg = "Failed to " + action + "."
        LOG.error(msg)
        raise


@contextlib.contextmanager
def log_and_reraise_smut_request_failed(action=None):
    """Catch SDK base exception and print error log before reraise exception.

    msg: the error message to be logged.
    """
    try:
        yield
    except exception.SDKSMUTRequestFailed as err:
        msg = ''
        if action is not None:
            msg = "Failed to %s. " % action
        msg += "SMUT error: %s" % err.format_message()
        LOG.error(msg)
        raise exception.SDKSMUTRequestFailed(err.results, msg)


@contextlib.contextmanager
def ignore_errors():
    """Only execute the clauses and ignore the results"""
    try:
        yield
    except Exception as err:
        msg = 'ignore an error:%s' % err.format_message()
        LOG.debug(msg)
        pass


def get_smut_userid():
    """Get the userid of smut server"""
    cmd = ["sudo", "/sbin/vmcp", "query userid"]
    try:
        userid = subprocess.check_output(cmd,
                                         close_fds=True,
                                         stderr=subprocess.STDOUT)
        userid = bytes.decode(userid)
        userid = userid.split()[0]
        return userid
    except Exception as err:
        msg = ("Could not find the userid of the smut server: %s") % err
        raise exception.SDKInternalError(msg=msg)


def get_namelist():
    """Generate namelist.

    Either through set CONF.zvm.namelist, or by generate based on smut userid.
    """
    if CONF.zvm.namelist is not None:
        # namelist length limit should be 64, but there's bug limit to 8
        # will change the limit to 8 once the bug fixed
        if len(CONF.zvm.namelist) <= 8:
            return CONF.zvm.namelist

    # return ''.join(('NL', get_smut_userid().rjust(6, '0')[-6:]))
    # py3 compatible changes
    userid = get_smut_userid()
    return 'NL' + userid.rjust(6, '0')[-6:]


def generate_iucv_authfile(fn, client):
    """Generate the iucv_authorized_userid file"""
    lines = ['#!/bin/bash\n',
             'echo -n %s > /etc/iucv_authorized_userid\n' % client]
    with open(fn, 'w') as f:
        f.writelines(lines)


@wrap_invalid_resp_data_error
def translate_response_to_dict(rawdata, dirt):
    """Translate SMUT response to a python dictionary.

    SMUT response example:
    keyword1: value1\n
    keyword2: value2\n
    ...
    keywordn: valuen\n

    Will return a python dictionary:
    {keyword1: value1,
     keyword2: value2,
     ...
     keywordn: valuen,}

    """
    data_list = rawdata.split("\n")
    data = {}

    for ls in data_list:
        for k in list(dirt.keys()):
            if ls.__contains__(dirt[k]):
                data[k] = ls[(ls.find(dirt[k]) + len(dirt[k])):].strip()
                break

    if data == {}:
        msg = ("Invalid smut response data. Error: No value matched with "
               "keywords. Raw Data: %(raw)s; Keywords: %(kws)s" %
               {'raw': rawdata, 'kws': str(dirt)})
        raise exception.SDKInternalError(msg=msg)

    return data


def make_dummy_image(image_path, d_type='CKD'):
    if d_type not in ('CKD', 'FBA'):
        d_type = 'CKD'

    d_unit = 'CYL'
    if d_type == 'FBA':
        d_unit = 'BLK'

    header = ("z/VM %(type)s Disk Image:           0 %(unit)s" %
              {'type': d_type, 'unit': d_unit})

    header = (' '.join((header, 'HLen: 0055', 'GZIP: 0')))
    with open(image_path, 'wb') as f:
        f.write(header.encode())


@contextlib.contextmanager
def acquire_lock(lock):
    """ lock wrapper """
    lock.acquire()
    try:
        yield
    finally:
        lock.release()


def check_userid_exist(userid):
    cmd = 'sudo vmcp q %s' % userid
    rc, output = execute(cmd)
    if re.search('(^HCP\w\w\w003E)', output):
        # userid not exist
        return False
    return True


def check_userid_on_others(userid):
    try:
        check_userid_exist(userid)
        cmd = 'sudo vmcp q %s' % userid
        rc, output = execute(cmd)
        if re.search(' - SSI', output):
            return True
        return False
    except Exception as err:
        msg = ("Could not find the userid: %s") % err
        raise exception.SDKInternalError(msg=msg)



# Copyright 2017 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import json
import mock
import requests
import unittest


from zvmconnector import restclient
from zvmsdk import config


CONF = config.CONF


class FakeResp(object):
    def __init__(self):
        self.content = '{"output": "{}"}'


class RESTClientTestCase(unittest.TestCase):
    """Testcases for RESTClient."""
    def setUp(self):
        self.client = restclient.RESTClient(ssl_enabled=False)
        self.fake_userid = 'userid01'
        self.base_url = 'http://127.0.0.1:8888'
        self.headers = {'Content-Type': 'application/json'}
        self.headers.update(self.headers or {})
        self.headers['X-Auth-Token'] = self._tmp_token()
        self.response = FakeResp()

        self.client_ssl = restclient.RESTClient(ssl_enabled=True)
        self.base_url_ssl = 'https://127.0.0.1:8888'

    def test_init_ComputeAPI(self):
        self.assertTrue(isinstance(self.client, restclient.RESTClient))

    def _tmp_token(self):
        token = '1234567890'
        return token

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_create(self, get_token, request):
        # method = 'POST'
        # url = '/guests'
        disks = [{'size': '3g', 'is_boot_disk': True}]
        # body = {'guest': {'userid': self.fake_userid,
        #                'vcpus': 1,
        #                'memory': 1024,
        #               'disk_list': disks,
        #               'user_profile': 'profile'
        #               }}
        # body = json.dumps(body)
        # header = self.headers
        # full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_create", self.fake_userid,
                         1, 1024, disk_list=disks,
                         user_profile='profile')
        # request.assert_called_with(method, full_uri,
        #                             data=body, headers=header,
        #                             verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_list(self, get_token, request):
        method = 'GET'
        url = '/guests'
        body = None
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_list")
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_inspect_stats(self, get_token, request):
        method = 'GET'
        url = '/guests/stats?userid=%s' % self.fake_userid
        body = None
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_inspect_stats", self.fake_userid)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

        fake_userid_string = 'userid1, userid2, userid3,userid4'
        url = '/guests/stats?userid=%s' % fake_userid_string
        full_uri = self.base_url + url
        self.client.call("guest_inspect_stats", fake_userid_string)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_inspect_vnics(self, get_token, request):
        method = 'GET'
        url = '/guests/interfacestats?userid=%s' % self.fake_userid
        body = None
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_inspect_vnics", self.fake_userid)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

        fake_userid_string = 'userid1, userid2, userid3,userid4'
        url = '/guests/interfacestats?userid=%s' % fake_userid_string
        full_uri = self.base_url + url
        self.client.call("guest_inspect_vnics", fake_userid_string)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guests_get_nic_info(self, get_token, request):
        method = 'GET'
        url = '/guests/nics?userid=%s&nic_id=%s&vswitch=%s' % (
                                                    self.fake_userid,
                                                    '1000',
                                                    'xcatvsw1')
        body = None
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guests_get_nic_info", userid=self.fake_userid,
                         nic_id='1000', vswitch='xcatvsw1')
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_delete(self, get_token, request):
        method = 'DELETE'
        url = '/guests/%s' % self.fake_userid
        body = None
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_delete", self.fake_userid)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_get_definition_info(self, get_token, request):
        method = 'GET'
        url = '/guests/%s' % self.fake_userid
        body = None
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_get_definition_info", self.fake_userid)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_start(self, get_token, request):
        method = 'POST'
        url = '/guests/%s/action' % self.fake_userid
        body = {'action': 'start'}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_start", self.fake_userid)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_stop(self, get_token, request):
        method = 'POST'
        url = '/guests/%s/action' % self.fake_userid
        body = {'action': 'stop'}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_stop", self.fake_userid)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_softstop(self, get_token, request):
        method = 'POST'
        url = '/guests/%s/action' % self.fake_userid
        body = {'action': 'softstop'}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_softstop", self.fake_userid)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_softstop_parameter_set_zero(self, get_token, request):
        method = 'POST'
        url = '/guests/%s/action' % self.fake_userid
        body = {'action': 'softstop', 'timeout': 0, 'poll_interval': 0}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_softstop", self.fake_userid,
                         timeout=0, poll_interval=0)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_pause(self, get_token, request):
        method = 'POST'
        url = '/guests/%s/action' % self.fake_userid
        body = {'action': 'pause'}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_pause", self.fake_userid)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_unpause(self, get_token, request):
        method = 'POST'
        url = '/guests/%s/action' % self.fake_userid
        body = {'action': 'unpause'}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_unpause", self.fake_userid)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_reboot(self, get_token, request):
        method = 'POST'
        url = '/guests/%s/action' % self.fake_userid
        body = {'action': 'reboot'}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_reboot", self.fake_userid)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_reset(self, get_token, request):
        method = 'POST'
        url = '/guests/%s/action' % self.fake_userid
        body = {'action': 'reset'}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_reset", self.fake_userid)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_get_console_output(self, get_token, request):
        method = 'POST'
        url = '/guests/%s/action' % self.fake_userid
        body = {'action': 'get_console_output'}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_get_console_output", self.fake_userid)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_capture(self, get_token, request):
        method = 'POST'
        url = '/guests/%s/action' % self.fake_userid
        body = {'action': 'capture',
                'image': 'image_captured'}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_capture", self.fake_userid,
                         'image_captured')
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_deploy(self, get_token, request):
        method = 'POST'
        url = '/guests/%s/action' % self.fake_userid
        body = {'action': 'capture',
                'image': 'image_captured'}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_capture", self.fake_userid,
                         'image_captured')
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_get_info(self, get_token, request):
        method = 'GET'
        url = '/guests/%s/info' % self.fake_userid
        body = None
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_get_info", self.fake_userid)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_get_info_ssl(self, get_token, request):
        method = 'GET'
        url = '/guests/%s/info' % self.fake_userid
        body = None
        header = self.headers
        full_uri = self.base_url_ssl + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client_ssl.call("guest_get_info", self.fake_userid)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_create_nic(self, get_token, request):
        # method = 'POST'
        # url = '/guests/%s/nic' % self.fake_userid
        body = {'nic': {'vdev': '123', 'nic_id': '1234',
                        'mac_addr': 'xx:xx:xx:xx:xx:xx',
                        'active': False}}
        body = json.dumps(body)
        # header = self.headers
        # full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_create_nic", self.fake_userid,
                         vdev='123', nic_id='1234',
                         mac_addr='xx:xx:xx:xx:xx:xx',
                         active=False)
        request.assert_called_once()
        # request.assert_called_with(method, full_uri,
        #                            data=body, headers=header,
        #                            verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_delete_nic(self, get_token, request):
        method = 'DELETE'
        url = '/guests/%s/nic/%s' % (self.fake_userid, '123')
        body = {'active': False}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_delete_nic", self.fake_userid, '123',
                         active=False)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_nic_couple_to_vswitch(self, get_token, request):
        method = 'PUT'
        url = '/guests/%s/nic/%s' % (self.fake_userid, '123')
        body = {'info': {'couple': True,
                         'vswitch': 'vswitch1',
                         'active': False}}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_nic_couple_to_vswitch", self.fake_userid,
                         '123', 'vswitch1', active=False)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_nic_uncouple_from_vswitch(self, get_token, request):
        method = 'PUT'
        url = '/guests/%s/nic/%s' % (self.fake_userid, '123')
        body = {'info': {'couple': False,
                         'active': False}}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_nic_uncouple_from_vswitch", self.fake_userid,
                         '123', active=False)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_create_network_interface(self, get_token, request):
        method = 'POST'
        networks = [{'ip_addr': '12.12.12.12'}]
        url = '/guests/%s/interface' % self.fake_userid
        body = {'interface': {'os_version': 'rhel7.2',
                              'guest_networks': networks,
                              'active': False}}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_create_network_interface", self.fake_userid,
                         'rhel7.2', networks, active=False)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_delete_network_interface(self, get_token, request):
        method = 'DELETE'
        url = '/guests/%s/interface' % self.fake_userid
        body = {'interface': {'os_version': 'rhel7.2',
                              'vdev': '123',
                              'active': False}}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_delete_network_interface", self.fake_userid,
                         'rhel7.2', '123', active=False)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_get_power_state(self, get_token, request):
        method = 'GET'
        url = '/guests/%s/power_state' % self.fake_userid
        body = None
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_get_power_state", self.fake_userid)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_create_disks(self, get_token, request):
        method = 'POST'
        disks = [{'size': '3g'}]
        url = '/guests/%s/disks' % self.fake_userid
        body = {'disk_info': {'disk_list': disks}}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_create_disks", self.fake_userid, disks)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_delete_disks(self, get_token, request):
        method = 'DELETE'
        vdevs = ['0101', '0102']
        url = '/guests/%s/disks' % self.fake_userid
        body = {'vdev_info': {'vdev_list': vdevs}}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_delete_disks", self.fake_userid, vdevs)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_config_minidisks(self, get_token, request):
        method = 'PUT'
        disks = [{'vdev': '0101', 'mntdir': '/mnt/0101'}]
        url = '/guests/%s/disks' % self.fake_userid
        body = {'disk_info': {'disk_list': disks}}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_config_minidisks", self.fake_userid, disks)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_host_get_info(self, get_token, request):
        method = 'GET'
        url = '/host'
        body = None
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("host_get_info")
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_host_diskpool_get_info(self, get_token, request):
        # wait host_diskpool_get_info bug fixed
        pass

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_image_import(self, get_token, request):
        method = 'POST'
        image_uri = 'file:///tmp/100.img'
        image_meta = {'os_version': 'rhel7.2', 'md5sum': 'dummy'}
        url = '/images'
        body = {'image': {'image_name': '100.img',
                          'url': image_uri,
                          'image_meta': image_meta}}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("image_import", '100.img',
                         image_uri, image_meta)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_image_delete(self, get_token, request):
        method = 'DELETE'
        url = '/images/%s' % '100.img'
        body = None
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("image_delete", '100.img')
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_image_export(self, get_token, request):
        method = 'PUT'
        destination = 'file:///tmp/export.img'
        url = '/images/%s' % '100.img'
        body = {'location': {'dest_url': destination}}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("image_export", '100.img',
                         destination)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_image_get_root_disk_size(self, get_token, request):
        method = 'GET'
        url = '/images/%s/root_disk_size' % '100.img'
        body = None
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("image_get_root_disk_size", '100.img')
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_token_create(self, get_token, request):
        method = 'POST'
        url = '/token'
        body = None
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("token_create")
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_vswitch_get_list(self, get_token, request):
        method = 'GET'
        url = '/vswitches'
        body = None
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("vswitch_get_list")
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_vswitch_create(self, get_token, request):
        method = 'POST'
        url = '/vswitches'
        body = {'vswitch': {'name': 'dummy'}}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("vswitch_create", 'dummy')
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_vswitch_delete(self, get_token, request):
        method = 'DELETE'
        url = '/vswitches/%s' % 'dummy'
        body = None
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("vswitch_delete", 'dummy')
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_vswitch_grant_user(self, get_token, request):
        method = 'PUT'
        url = '/vswitches/%s' % 'dummy'
        body = {'vswitch': {'grant_userid': self.fake_userid}}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("vswitch_grant_user", 'dummy', self.fake_userid)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_vswitch_revoke_user(self, get_token, request):
        method = 'PUT'
        url = '/vswitches/%s' % 'dummy'
        body = {'vswitch': {'revoke_userid': self.fake_userid}}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("vswitch_revoke_user", 'dummy', self.fake_userid)
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_vswitch_set_vlan_id_for_user(self, get_token, request):
        method = 'PUT'
        url = '/vswitches/%s' % 'dummy'
        body = {'vswitch': {'user_vlan_id': {'userid': self.fake_userid,
                                             'vlanid': 'vlan_id'}}}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("vswitch_set_vlan_id_for_user",
                         'dummy', self.fake_userid, 'vlan_id')
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_resize_mem(self, get_token, request):
        method = 'POST'
        url = '/guests/%s/action' % self.fake_userid
        body = {'action': 'resize_mem',
                'size': '4g'}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_resize_mem", self.fake_userid,
                         '4g')
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)

    @mock.patch.object(requests, 'request')
    @mock.patch('zvmconnector.restclient.RESTClient._get_token')
    def test_guest_live_resize_mem(self, get_token, request):
        method = 'POST'
        url = '/guests/%s/action' % self.fake_userid
        body = {'action': 'live_resize_mem',
                'size': '4g'}
        body = json.dumps(body)
        header = self.headers
        full_uri = self.base_url + url
        request.return_value = self.response
        get_token.return_value = self._tmp_token()

        self.client.call("guest_live_resize_mem", self.fake_userid,
                         '4g')
        request.assert_called_with(method, full_uri,
                                   data=body, headers=header,
                                   verify=False)



# Copyright 2017,2018 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Handler for the root of the sdk API."""

import json

from zvmsdk.sdkwsgi import util
from zvmsdk.sdkwsgi.handlers import tokens
from zvmsdk import utils
from zvmsdk import version as sdk_version


APIVERSION = '1.0'


@util.SdkWsgify
@tokens.validate
def version(req):
    min_version = APIVERSION
    max_version = APIVERSION

    version_data = {
        'version': sdk_version.__version__,
        'api_version': '%s' % APIVERSION,
        'max_version': max_version,
        'min_version': min_version,
    }
    output = {
        "rs": 0,
        "overallRC": 0,
        "modID": None,
        "rc": 0,
        "errmsg": "",
        "output": version_data
    }
    info_json = json.dumps(output)
    req.response.body = utils.to_utf8(info_json)
    req.response.content_type = 'application/json'
    req.response.status = 200

    return req.response



# Messages for Systems Management Ultra Thin Layer
#
# Copyright 2017 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""
List of modules that use these messages and their module id.
ModId  Module
-----  --------------------
CVM    changeVM.py
CMD    cmdVM.py
DVM    deleteVM.py
GUT    generalUtils.py
GHO    getHost.py
GVM    getVM.py
MVM    makeVM.py
MIG    migrateVM.py
PVM    powerVM.py
RQH    ReqHandle.py
SMC    Reserved for smcli
SMP    smapi.py
VMU    vmUtils.py

List of messages.  Message id is the key.
Messages are grouped by their message number.
Each message is defined in a array that provides:
    - Dictionary of values for the result structure,
      e.g. {'overallRC': 4, 'rc': 100}
      The keys in the dictionary can contain: 'overallRC', 'rc', 'rs'
      'errno', and 'strError'.  These allow us to use the dictionary to
      update the ReqHandle results dictionary.
    - Message text (may contain substitution directives, e.g. %s, %i)
    - Sample text in a python tuple
"""
msg = {
    # 0001-0099: Parsing Messages
    '0001': [{'overallRC': 4, 'rc': 4, 'rs': 1},
            "ULT%s0001E %s %s subfunction's operand at position %i (%s) " +
            "is not an integer: %s",
            ('RQH', 'FUNCTION_NAME', 'SUBFUNCTION_NAME', 'OPERAND_POSITION',
             'OPERAND', 'OPERAND_VALUE')],
        # Explain: An error was detected while parsing the command.
        #   The indicated operand is not an integer.
        # SysAct: Processing of the subfunction terminates.
        # UserResp: Correct the syntax so that the indicated operand is
        #   an integer, e.g., 10 and reissue the command.
    '0002': [{'overallRC': 4, 'rc': 4, 'rs': 2},
            "ULT%s0002E %s's %s subfunction is missing positional " +
            "operand (%s) at position %i.",
            ('RQH', 'FUNCTION_NAME', 'SUBFUNCTION_NAME', 'OPERAND',
             'OPERAND_POSITION')],
        # Explain: An error was detected while parsing the command.
        #   A positional operand is missing.
        # SysAct: Processing of the subfunction terminates.
        # UserResp: Correct the syntax by specifying the missing operand
        #   and reissue the command.
    '0003': [{'overallRC': 4, 'rc': 4, 'rs': 3},
            "ULT%s0003E %s's %s subfunction %s keyword operand is " +
            "missing a value.",
            ('RQH', 'FUNCTION_NAME', 'SUBFUNCTION_NAME', 'KEYWORD')],
        # Explain: An error was detected while parsing the command.
        #   A keyword operand that requires a value was specified without
        #   the value.
        # SysAct: Processing of the subfunction terminates.
        # UserResp: Correct the syntax to provide a value for the specified
        #   keyword operand and reissue the command.
    '0004': [{'overallRC': 4, 'rc': 4, 'rs': 4},
            "ULT%s0004E %s's %s subfunction %s keyword operand is not " +
            "an integer: %s",
            ('RQH', 'FUNCTION_NAME', 'SUBFUNCTION_NAME', 'KEYWORD',
             'KEYWORD_VALUE')],
        # Explain: An error was detected while parsing the command.
        #   The specified operand for a keyword is not an integer.
        # SysAct: Processing of the subfunction terminates.
        # UserResp: Correct the syntax so that the keyword operand is
        #   an integer, e.g., 10 and reissue the command.
    '0005': [{'overallRC': 4, 'rc': 4, 'rs': 5},
            "ULT%s0005E %s's %s subfunction does not recognize keyword: %s",
            ('RQH', 'FUNCTION_NAME', 'SUBFUNCTION_NAME', 'KEYWORD')],
        # Explain: An error was detected while parsing the command.
        #   An unrecognized keyword was encountered.
        # SysAct: Processing of the subfunction terminates.
        # UserResp: Correct the syntax to specify a recognized keyword
        #   and reissue the command.
    '0006': [{'overallRC': 4, 'rc': 4, 'rs': 6},
            "ULT%s0006E %s's %s subfunction encountered an unknown " +
            "operand: %s",
            ('RQH', 'FUNCTION_NAME', 'SUBFUNCTION_NAME', 'OPERAND')],
        # Explain:  An error was detected while parsing the command.
        #   An unknown operand was encountered.
        # SysAct: Processing of the subfunction terminates.
        # UserResp: Correct the syntax and reissue the command.
    '0007': [{'overallRC': 4, 'rc': 4, 'rs': 7},
            "ULT%s0007E Unrecognized function: %s",
            ('RQH', 'FUNCTION_NAME')],
        # Explain:  An error was detected while parsing the command.
        #   The specified function is not known.
        # SysAct: Processing of the subfunction terminates.
        # UserResp: Correct the syntax and reissue the command.
    '0008': [{'overallRC': 4, 'rc': 4, 'rs': 8},
            "ULT%s0008E Specified function is not 'HELP' or 'VERSION': %s",
            ('RQH', 'SPECIFIED_FUNCTION')],
        # Explain: An error was detected while parsing the command.
        #   The specified function was not 'HELP' or 'VERSION' which are the
        #   only valid functions for a command of the specified length.
        # SysAct: Processing of the function terminates.
        # UserResp: Correct the syntax and reissue the command.
    '0009': [{'overallRC': 4, 'rc': 4, 'rs': 9},
            "ULT%s0009E Too few arguments specified.",
            ('RQH')],
        # Explain: An error was detected while parsing the command.
        #   The minimum number of arguments were not provided for the command.
        # SysAct: Processing of the subfunction terminates.
        # UserResp: Correct the syntax and reissue the command.
    '0010': [{'overallRC': 4, 'rc': 4, 'rs': 10},
            "ULT%s0010E Userid is missing",
            ('RQH')],
        # Explain: An error was detected while parsing the command.
        #   A userid operand was not specified.
        # SysAct: Processing of the subfunction terminates.
        # UserResp: Correct the syntax and specify the userid along
        #   with any other required operands and reissue the command.
    '0011': [{'overallRC': 4, 'rc': 4, 'rs': 11},
            "ULT%s0011E Subfunction is missing. It should be one of " +
            "the following: %s",
            ('RQH', 'VALID_SUBFUNCTIONS')],
        # Explain: An error was detected while parsing the command.
        #   The name of the subfunction was not specified.
        # SysAct: Processing of the function terminates.
        # UserResp: Correct the syntax and specify the userid along
        #   with any other required operands and reissue the command.
    '0012': [{'overallRC': 4, 'rc': 4, 'rs': 12},
            "ULT%s0012E The request data is not one of the supported types " +
            "of list or string: %s",
            ('RQH', 'REQUEST_DATA_TYPE')],
        # Explain: The ReqHandle parseCmdline method was called with
        #   the request passed in a variable that was not a
        #   list or base string.  Only these types of variables are
        #   supported for passing of the request to be parsed.
        # SysAct: Processing of the function terminates.
        # UserResp: Correct the calling function to use either a
        #   list or base string to hold the request to be processed
        #   and reinvoke the call.
    '0013': [{'overallRC': 4, 'rc': 4, 'rs': 13},
            "ULT%s0010E The desired state was: %s. Valid states are: %s",
            ('RQH', 'DESIRED_STATS', 'VALID_STATS')],
        # Explain: An error was detected while parsing the command.
        #   The state operand value is not one of the accepted values.
        #   The valid values are shown in the message.
        # SysAct: Processing of the subfunction terminates.
        # UserResp: Correct the syntax to use one of the valid states
        #   and reissue the command.
    '0014': [{'overallRC': 4, 'rc': 4, 'rs': 14},
            "ULT%s0014E The option %s was specified but the option %s " +
            "was not specified.  These options must both be specified.",
            ('RQH', 'OPTION1', 'OPTION2')],
        # Explain: An error was detected while parsing the command.
        #   An option was specified which required a related
        #   option that was not specified.
        # SysAct: Processing of the subfunction terminates.
        # UserResp: Correct the syntax to specify both options and
        #   reissue the command.
    '0015': [{'overallRC': 4, 'rc': 4, 'rs': 15},
            "ULT%s0015E The file system was not 'ext2', 'ext3', " +
            "'ext4', 'xfs' or 'swap': %s",
            ('RQH', 'FILE_SYSTEM')],
        # Explain: An error was detected while parsing the command.
        #   The type of file system does not match one of the valid
        #   values.
        # SysAct: Processing of the subfunction terminates.
        # UserResp: Correct the syntax to use a valid file system type
        #   and reissue the command.
    '0016': [{'overallRC': 4, 'rc': 4, 'rs': 16},
            "ULT%s0016E The scp Data Type was not 'hex', 'ebcdic', " +
            "or 'delete': %s",
            ('RQH', 'DATA_TYPE')],
        # Explain: An error was detected while parsing the command.
        #   The value specified for the scp data type is not one of the
        #   recognized values.
        # SysAct: Processing of the subfunction terminates.
        # UserResp: Correct the syntax to use a valid scp data type and
        #   reissue the command.
    '0017': [{'overallRC': 4, 'rc': 4, 'rs': 17},    # dict is not used
            "ULT%s0017W The maxwait time %i sec is not evenly divisible " +
            "by the poll interval %i sec.  Maximum wait time will be %i " +
            "sec or %i poll intervals.",
            ('RQH', 'MAX_WAIT', 'POLL_INTERVAL', 'RECOMMEND_MAX_WAIT',
             'RECOMMEND_POLL_INTERVAL')],
        # Explain: When trying to determine how many polling intervals
        #   to wait for a desired guest power state, it was found that the
        #   specified maximum wait time was not evenly divisible by the
        #   number of polling interval seconds.  The program instead
        #   rounded the maximum wait time up to be evenly divisble
        #   by the polling interval.
        # SysAct: Processing of the function continues with the
        #   new wait time.
        # UserResp: If the wait time is unacceptably long, invoke
        #   the function with a maximum wait time and polling
        #   interval time which are evenly divisible and of an
        #   acceptable duration.

    # 0200-0299: Utility Messages
    '0200': [{'overallRC': 4, 'rc': 4, 'rs': 200},
            "ULT%s0200E The size of the disk is not valid: %s",
            ('GUT', 'DISK_SIZE')],
        # Explain: An error was encountered while attempting
        #   to convert the size of a disk from bytes to cylinders
        #   (for 3390 type disks) or bytes to blocks (for FBA type disks).
        #   This error can be caused by specifying the size as only a
        #   magnitude, (e.g., 'G' or 'M') instead of an integer
        #   appended to a magnitude (e.g., '20G').
        # SysAct: Processing of the subfunction terminates.
        # UserResp: Correct the disk size to specify a disk magnitude
        #   that includes the integer portion of the size in addition
        #   to the magnitude and reissue the command.
    '0201': [{'overallRC': 4, 'rc': 4, 'rs': 201},
            "ULT%s0201E Failed to convert %s to a number of blocks.",
            ('GUT', 'DISK_SIZE')],
        # Explain: An error was encountered while attempting
        #   to convert the size of a disk from bytes to blocks.
        #   The size ended with a magnitude character and should have
        #   had an integer value prepended to the magnitude character
        #   (e.g. '10M' or '10G').
        #   The probable cause of the error is that the integer
        #   portion of the size contains a non-numeric character.
        # SysAct: Processing of the subfunction terminates.
        # UserResp: Correct the disk size to specify a valid value
        #   and reissue the command.
    '0202': [{'overallRC': 4, 'rc': 4, 'rs': 202},
            "ULT%s0202E %s is not an integer size of blocks.",
            ('GUT', 'NUM_BLOCKS')],
        # Explain: An error was encountered while attempting
        #   to convert the size of a disk from bytes to blocks.
        #   The size did not end with a valid magnitude character
        #   (i.e., 'M' or 'G') so it was treated as an integer
        #   value (e.g. '100000').  The probable cause of this
        #   error is that the size contains non-numeric
        #   characters.
        # SysAct: Processing of the subfunction terminates.
        # UserResp: Correct the disk size to specify a valid value
        #   and reissue the command.
    '0203': [{'overallRC': 4, 'rc': 4, 'rs': 203},
            "ULT%s0203E Failed to convert %s to a number of cylinders.",
            ('GUT', 'DISK_SIZE')],
        # Explain: An error was encountered while attempting
        #   to convert the size of a disk from bytes to cylinders.
        #   The size ended with a magnitude character and should have
        #   had an integer value prepended to the magnitude character
        #   (e.g. '10M' or '10G').
        #   The probable cause of the error is that the integer
        #   portion of the size contains non-numeric characters.
        # SysAct: Processing of the subfunction terminates.
        # UserResp: Correct the disk size to specify a valid value
        #   and reissue the command.
    '0204': [{'overallRC': 4, 'rc': 4, 'rs': 204},
            "ULT%s0204E %s is not an integer size of cylinders.",
            ('GUT', 'DISK_SIZE')],
        # Explain: An error was encountered while attempting
        #   to convert the size of a disk from bytes to cylinders.
        #   The size did not end with a valid magnitude character
        #   (i.e., 'M' or 'G') so it was treated as an integer
        #   value (e.g. '100000').  The probable cause of this
        #   error is that the size contains non-numeric
        #   characters.
        # SysAct: Processing of the subfunction terminates.
        # UserResp: Correct the disk size to specify a valid value
        #   and reissue the command.
    '0205': [{'overallRC': 4, 'rc': 4, 'rs': 205},
            "ULT%s0204E memory size did not end with suffix 'G' or 'M'.",
            ('MVM')],
        # Explain: An error was encountered while handling memory size.
        #   The size did not end with a valid magnitude character
        #   (i.e., 'M' or 'G').
        # SysAct: Processing of the subfunction terminates.
        # UserResp: Correct the memory size to specify a valid value
        #   and reissue the command.
    '0206': [{'overallRC': 4, 'rc': 4, 'rs': 206},
            "ULT%s0204E Max memory size %s specified is less than " +
            "initial memory size %s.",
            ('MVM', 'MAX_MEM_SIZE', 'INIT_MEM_SIZE')],
        # Explain: An error was encountered while handling memory size.
        #   The size did not end with a valid magnitude character
        #   (i.e., 'M' or 'G').
        # SysAct: Processing of the subfunction terminates.
        # UserResp: Correct the memory size to specify a valid value
        #   and reissue the command.
    # 0207-0299: Available

    # SMCLI and SMAPI related messages.
    '0300': [{'overallRC': 8},    # dict is not used.
            "ULT%s0300E SMAPI API failed: %s, overall rc: %s, rc: %s, " +
            "rs: %s, errno: %s, cmd: %s, out: %s",
            ('SMP', 'API_NAME', 'OVERALLRC', 'RC', 'RS', 'ERRNO', 'CMD',
             'OUTPUT')],
        # Explain: The smcli program was invoked to call z/VM SMAPI for
        #   the indicated API.  An error was encountered.  The overall rc
        #   indicates the location of the problem:
        #   8  - SMAPI returned the error.  The rc and rs values are
        #        the values provided by SMAPI.  The z/VM Systems
        #        Management Application Programming book provides
        #        additional documentation related to the return code and
        #        reason code in the API description and in the "Return
        #        and Reason Code Summary" chapter.
        #   24 - The smcli program identified a parameter validation
        #        error.  A message will indicate what was detected.
        #        It could be a missing parameter, invalid parameter, etc.
        #        Invoke the smcli program using the -h parameter and the
        #        API name shown in the error message to obtain additional
        #        invocation help, e.g. "./smcli Image_Query_DM -h".
        #        In addition, the z/VM Systems Management Application
        #        Programming book provides additional documentation
        #        related to the return code and reason code in the API
        #        description.
        #   25 - The smcli program encountered an internal error.
        #        The rc and errno contains information related to the
        #        problem.  The error message from the smcli invocation
        #        and log entries in the system log provide the most
        #        useful information to debug this error.
        # SysAct: Processing of the function terminates.
        # UserResp: Determine the cause of the problem using the
        #   information described in the explanation section.  Reinvoke
        #   the function after you correct the problem.
    '0301': [{'overallRC': 25, 'rc': 301, 'rs': 0},
            "ULT%s0301E SMAPI API failed: %s, response header does not " +
            "have the expected 3 values before the (details) string. " +
            "cmd: %s, response header: %s, out: %s",
            ('SMP', 'API_NAME', 'CMD', 'HEADER', 'OUTPUT')],
        # Explain: The smcli program was invoked to call z/VM SMAPI for
        #   the indicated API.  The expected response from the smcli
        #   program has a header that contains 3 integers followed by
        #   the string '(details)'.  The response returned by the program
        #   does not have the expected header.  This indicates a problem
        #   in the smcli program or a problem invoking the smcli program.
        # SysAct: Processing of the function terminates.
        # UserResp: Determine the cause of the failure.  If it is not a
        #   Linux permission problem then investigate a possible coding
        #   error in the smcli program.  Correct the problem and reinvoke
        #   the function.
    '0302': [{'overallRC': 25, 'rc': 302, 'rs': 0},
            "ULT%s0302E SMAPI API failed: %s, word 1 in " +
            "the response header is not an integer or in the range of " +
            "expected values. word 1: %s, cmd: %s, response " +
            "header: %s, out: %s",
            ('SMP', 'API_NAME', 'WORD1', 'CMD', 'HEADER', 'OUTPUT')],
        # Explain: The smcli program was invoked to call z/VM SMAPI for
        #   the indicated API.  The expected response from the smcli
        #   program has a header that contains 3 integers followed by
        #   the string '(details)'.  The first word should provide the
        #   overall return code of the smcli invocation that indicates
        #   where the failure occurred.  However, it does not represent
        #   an integer value or is not the expected error values of
        #   8, 24 or 25.  This indicates a problem in the smcli program.
        # SysAct: Processing of the function terminates.
        # UserResp: Determine the cause of the failure.
        #   Investigate a possible coding error in the smcli program.
        #   Correct the problem and reinvoke the function.
    '0303': [{'overallRC': 25, 'rc': 303, 'rs': 0},
            "ULT%s0303E SMAPI API failed: %s, word 2 in the response " +
            "header is not an integer. word 2: %s, cmd: %s, response " +
            "header: %s, out: %s",
            ('SMP', 'API_NAME', 'WORD2', 'CMD', 'HEADER', 'OUTPUT')],
        # Explain: The smcli program was invoked to call z/VM SMAPI for
        #   the indicated API.  The expected response from the smcli
        #   program has a header that contains 3 integers followed by
        #   the string '(details)'.  The second word should provide the
        #   specific return code of the smcli invocation.  However, it
        #   does not represent an integer value.  This indicates a
        #   problem in the smcli program.
        # SysAct: Processing of the function terminates.
        # UserResp: Determine the cause of the failure.
        #   You may need an update to the smcli program.
        #   Correct the problem and reinvoke the function.
    '0304': [{'overallRC': 25, 'rc': 304, 'rs': 0},
            "ULT%s0304E SMAPI API failed: %s, word 3 in the response " +
            "header is not an integer. word 3: %s, cmd: %s, response " +
            "header: %s, out: %s",
            ('SMP', 'API_NAME', 'WORD3', 'CMD', 'HEADER', 'OUTPUT')],
        # Explain: The smcli program was invoked to call z/VM SMAPI for
        #   the indicated API.  The expected response from the smcli
        #   program has a header that contains 3 integers followed by
        #   the string '(details)'.  The third word should provide
        #   the reason code or errno, depending on the error.  However,
        #   it does not represent an integer value.  This indicates
        #   a problem in the smcli program.
        # SysAct: Processing of the function terminates.
        # UserResp: Determine the cause of the failure.
        #   You may need an update to the smcli program.
        #   Correct the problem and reinvoke the function.
    '0305': [{'overallRC': 99, 'rc': 305, 'rs': 0},
            "ULT%s0305E Exception received on an attempt to " +
            "communicate with SMAPI, cmd: %s, exception: %s, " +
            "details: %s",
            ('SMP', 'CMD', 'EXCEPTION', 'EXCEPTION_DETAILS')],
        # Explain: The function attempted to invoke the smcli
        #   program to communicate with z/VM SMAPI.  This failed
        #   due to the exception shown in the message.
        # SysAct: Processing of the function terminates.
        # UserResp: Determine the cause of the failure using
        #   the exception and exception details provided in the message.
        #   Reinvoke the function after correcting the problem.
    # 0306-0310: Available

    # IUCV related messages
    '0311': [{'overallRC': 2, 'rc': 2, 'rs': 99},    # dict is not used.
            "ULT%s0311E On %s, command sent through IUCV failed, " +
            "rc in response string is not an integer. " +
            "cmd: %s, rc: %s, out: %s",
            ('SMP', 'USERID', 'CMD', 'RC', 'OUTPUT')],
        # Explain: The IUCV client returned a non-integer return
        #   code value.
        # SysAct: Processing of the function terminates.
        # UserResp: Contact the support team with the information
        #   included in the message.  Investigate the problem in the
        #   IUCVCLNT program, fix the code and reinvoke the function.
    '0312': [{'overallRC': 2, 'rc': 2, 'rs': 99},    # dict is not used.
            "ULT%s0312E On %s, command sent through IUCV failed, " +
            "reason code in response string is not an integer. " +
            "cmd: %s, rc: %s, rs: %s, out: %s",
            ('SMP', 'USERID', 'CMD', 'RC', 'RS', 'OUTPUT')],
        # Explain: The IUCV client returned a non-integer reason
        #   code value.
        # SysAct: Processing of the function terminates.
        # UserResp: Contact the support team with the information
        #   included in the message.  The IUCVCLNT program is the probable
        #   cause of the failure.  This will require a code change.
    '0313': [{'overallRC': 2, 'rc': 1},      # dict is not used.
            "ULT%s0313E On %s, command sent through IUCV was not " +
            "authorized or a generic Linux error occurred. " +
            "cmd: %s, rc: %s, rs: %s, out: %s",
            ('SMP', 'USERID', 'CMD', 'RC', 'RS', 'OUTPUT')],
        # Explain: The command that was sent to the target system failed.
        #   The cause of the failure is either a Linux permission problem
        #   for the command being executed or a generic Linux error.
        # SysAct: Processing of the function terminates.
        # UserResp: Use the information included in the message to determine
        #   the cause of the failure on the target system and correct the
        #   problem.  After correcting the problem, you should be able to
        #   reinvoke the failing function.
    '0314': [{'overallRC': 2, 'rc': 2},      # dict is not used.
            "ULT%s0314E IUCV client parameter error sending command to %s. " +
            "cmd: %s, rc: %s, rs: %s, out: %s",
            ('SMP', 'USERID', 'CMD', 'RC', 'RS', 'OUTPUT')],
        # Explain: The IUCVCLNT program communicates with managed
        #   systems using IUCV.  The program detected invocation
        #   errors.  This can be caused by a problem in the level of the
        #   IUCVCLNT program or the function that invoked it.
        # SysAct: Processing of the function terminates.
        # UserResp: Use the information included in the message to determine
        #   the cause of the failure.  This could require the support
        #   team to provide a code change to either the IUCVCLNT program
        #   or the code that invoked it.
    '0315': [{'overallRC': 2, 'rc': 4},      # dict is not used.
            "ULT%s0315E IUCV socket error sending command to %s. " +
            "cmd: %s, rc: %s, rs: %s, out: %s",
            ('SMP', 'USERID', 'CMD', 'RC', 'RS', 'OUTPUT')],
        # Explain: The IUCVCLNT program communicates with managed
        #   systems using IUCV.  The program encountered an IUCV
        #   communication failure when it attempted to send a
        #   command to the managed system.
        #   This is probably caused by a failure in the managed system
        #   that prevents the system from receiving the command.
        #   One cause could be that the system logged off z/VM.
        #   Another cause is that the managed system is not running the
        #   related IUCV daemon or has not authorized access by
        #   the system contacting it in the /etc/iucv_authorized_userid
        #   file.
        # SysAct: Processing of the function terminates.
        # UserResp: Use the information included in the message to
        #   determine the cause of the failure.  Reinvoke the function
        #   after you correct the problem.
    '0316': [{'overallRC': 2, 'rc': 8},      # dict is not used.
            "ULT%s0316E On %s, command sent through IUCV failed. " +
            "cmd: %s, rc: %s, rs: %s, out: %s",
            ('SMP', 'USERID', 'CMD', 'RC', 'RS', 'OUTPUT')],
        # Explain: The command that was sent to the target system failed.
        # SysAct: Processing of the function terminates.
        # UserResp: Use the information included in the message to
        #   determine the cause of the failure.  Reinvoke the function
        #   after you correct the problem.
    '0317': [{'overallRC': 2, 'rc': 16},     # dict is not used.
            "ULT%s0317E File transport failure while processing " +
            "command for %s. " +
            "cmd: %s, rc: %s, rs: %s, out: %s",
            ('SMP', 'USERID', 'CMD', 'RC', 'RS', 'OUTPUT')],
        # Explain: The IUCVCLNT program failed to send a file to
        #   the target system.
        # SysAct: Processing of the function terminates.
        # UserResp: Use the information included in the message to
        #   determine the cause of the failure.  Reinvoke the function
        #   after you correct the problem.
    '0318': [{'overallRC': 2, 'rc': 32},     # dict is not used.
            "ULT%s0318E On %s, IUCV server file was not found. " +
            "cmd: %s, rc: %s, rs: %s, out: %s",
            ('SMP', 'USERID', 'CMD', 'RC', 'RS', 'OUTPUT')],
        # Explain: The IUCVCLNT program failed to find the IUCVSERV
        #   file on the local system.  This file is expected to exist
        #   in the same directory as the IUCVCLNT program.
        # SysAct: Processing of the function terminates.
        # UserResp: Determine the reason that the IUCVSERV file could
        #   not be located and correct the problem.  Reinvoke the
        #   function after you correct the problem.
    '0319': [{'overallRC': 2},               # dict is not used.
            "ULT%s0319E Unrecognized IUCV client error encountered " +
            "while sending a command through IUCV to %s. " +
            "cmd: %s, rc: %s, rs: %s, out: %s",
            ('SMP', 'USERID', 'CMD', 'RC', 'RS', 'OUTPUT')],
        # Explain: The IUCVCLNT program returned a non-zero return code
        #   that does not correspond to a recognized error value.
        # SysAct: Processing of the function terminates.
        # UserResp: Determine the cause of the error using the
        #   information in the message.  Log files on the local system
        #   and the target system may contain useful information to
        #   identify the failure.  Reinvoke the function after you
        #   correct the problem.

    # General subfunction processing messages
    '0400': [{'overallRC': 4, 'rc': 4, 'rs': 400},
            "ULT%s0400E The worker script %s does not exist.",
            ('GUT', 'SCRIPT_NAME')],
        # Explain: The activation engine modification script specified
        #   for "aeScript" cannot be found.
        # SysAct: Processing of the function ends with no action
        #   taken.
        # UserResp: Correct the function call to point to an existing script
        #   and reinvoke the function.
    '0401': [{'overallRC': 4, 'rc': 7, 'rs': 401},
            "ULT%s0401E Failed to punch %s file to guest: %s, out: %s",
            ('GUT', 'FILE_LOCATION', 'USERID', 'OUTPUT')],
        # Explain: The vmur punch command failed for the specified
        #   reason.
        # SysAct: Processing of the function ends with no action
        #   taken.
        # UserResp: Look up the reason the vmur command failed, correct
        #   the problem and reinvoke the function.
    '0402': [{'overallRC': 4, 'rc': 5, 'rs': 402},
            "ULT%s0402E No information was found for the specified " +
            "pool(s): %s",
            ('GUT', 'DISK_POOL')],
        # Explain: Image_Volume_Space_Query_DM returned successfully
        #   but the list of pools of the specified names was empty.
        # SysAct: Processing terminates with an error.
        # UserResp:  Correct the function call to query existing pools and
        #   reinvoke the function.
    '0403': [{'overallRC': 4, 'rc': 99, 'rs': 403},     # dict is not used.
            "ULT%s0403E  Failed to purge reader file %s, out: %s",
            ('GUT', 'SPOOL_ID', 'OUTPUT')],
        # Explain: The vmcp purge reader file command failed.
        #   The system was already in the process of cleaning up from a
        #   failed attempt to punch a file, so the error processing
        #   continues.
        # SysAct: Error processing continues.
        # UserResp: Manually clean up the specified reader file using
        #   CP commands to avoid problems with old files or spool space
        #   filling up.
    '0404': [{'overallRC': 4, 'rc': 8, 'rs': 404},
            "ULT%s0404E Failed to spool the punch to the specified class %s" +
            ", out:%s ",
            ('GUT', 'SPOOL_CLASS', 'OUTPUT')],
        # Explain: The vmcp change reader command failed with the
        #   specified output.
        # SysAct: Processing of the function ends with no action
        #   taken.
        # UserResp: Look up the reason the change reader command failed
        #   in the CP messages book or vmcp help.  Correct the problem
        #   and reinvoke the function.
    '0405': [{'overallRC': 4, 'rc': 6, 'rs': 405},
            "ULT%s0405E Unable to obtain information related to: " +
            "%s. Command used was: %s. Output was: %s",
            ('GUT', 'KEYWORD', 'CMD', 'OUTPUT')],
        # Explain: While gathering hypervisor information, one of the
        #   commands used failed and that piece of information could
        #   not be queried.
        # SysAct: The getHost GENERAL function returns "no info" for
        #   the specified hypervisor information.
        # UserResp: If the information is needed, investigate the
        #   failure, correct it and reinvoke the function.
    '0406': [{'overallRC': 4, 'rc': 9, 'rs': 406},
            "ULT%s0406E Failed to punch %s because of VMUR timeout ",
            ('GUT', 'FILE_LOCATION')],
        # Explain: When punching a file to the reader, the vmur punch
        #   command is issued up to 5 times with increasing timeouts.
        #   This error comes after the 5th try if the vmur command
        #   was still unsuccessful.
        # SysAct:  Processing of the function ends with no action taken.
        # UserResp: This error could be because of another process
        #   also issuing vmur commands at the same time.  Wait a few
        #   seconds and reinvoke the function.
    '0407': [{'overallRC': 4, 'rc': 4, 'rs': 407},     # dict is not used.
            "ULT%s0407W Unable to spool reader to all classes, " +
            "it is possible that there may be additional console " +
            "files available that are not listed in the response. " +
            "Response from %s is %s",
            ('GUT', 'CMD', 'OUTPUT')],
        # Explain: The vmcp spool reader class * command was not
        #   successful.  This means the reader could not be changed
        #   to get files of all classes, and thus there could be
        #   files that are ignored.
        # SysAct: Processing of the function continues.
        # UserResp: If missing files are suspected, investigate the
        #   cause of the failure in the CP messages book or vmcp
        #   help and reinvoke the function.
    '0408': [{'overallRC': 4, 'rc': 4, 'rs': 408},
            "ULT%s0408E Error getting list of files in the reader " +
            "to search for logs from user %s. Response from %s is %s",
            ('GUT', 'USERID', 'CMD', 'OUTPUT')],
        # Explain: The vmur list command failed.  The list of files
        #   in the user's reader could not be determined.
        # SysAct:  Processing of the function ends with no action taken.
        # UserResp: Investigate the failure in vmur and correct the
        #   problem, then reinvoke the function.
    '0409': [{'overallRC': 4, 'rc': 4, 'rs': 409},
            "ULT%s0409E Unable to get console log for user %s. " +
            "The userid is either: not logged on, not spooling " +
            "its console, or has not created any console output. " +
            "Error rc=rs=8 returned from " +
            "Image_Console_Get.",
            ('GUT', 'USERID')],
        # Explain: The Image_Console_Get SMAPI call returned that
        #   there were no spool files available for that user.
        # SysAct: Processing of the function ends with no action taken.
        # UserResp: Check that the user is logged on, has issued a
        #   SPOOL CONSOLE command and has done some actions that
        #   would result in console output, then reinvoke the function.
    '0410': [{'overallRC': 4, 'rc': 4, 'rs': 410},
            "ULT%s0410E Unable to get console log for user %s " +
            "no spool files were found in our reader from this " +
            "user, it is possible another process has already " +
            "received them.",
            ('GUT', 'USERID')],
        # Explain: The Image_Console_Get SMAPI call should have
        #   put files of class T and "CON" with the userid as the
        #   filename in our reader.  However no files were found
        #   in the vmur list output with these characteristcs.
        # SysAct: Processing of the function ends with no action taken.
        # UserResp: Likely another process in this virtual machine
        #   has already processed the spool files. They are gone.
    '0411': [{'overallRC': 4, 'rc': 4, 'rs': 411},
            "ULT%s0411E Unable to receive console output file. " +
            "Reader not online.  /sys/bus/ccw/drivers/vmur/0.0.000c" +
            "/online = 0",
            ('GUT')],
        # Explain: The reader is typically at virtual device address
        #   x'000C'.  Linux does not believe this device is online.
        # SysAct:  Processing of the function ends with no action taken.
        # UserResp: If the reader is at a different virtual device
        #   address, update the SMUT code to recognize the alternative
        #   device address, otherwise bring the reader at x'000C' online
        #   to Linux.  Then, reinvoke the function.
    '0412': [{'overallRC': 4, 'rc': 4, 'rs': 412},     # dict is not used.
            "ULT%s0412E Malformed reply from SMAPI, unable to fill " +
            "in performance information, exception: %s, " +
            "details: %s, Response: %s",
            ('GUT', 'EXCEPTION', 'EXCEPTION_DETAILS', 'OUTPUT')],
        # Explain: An error was encountered while processing the
        #   response information from the SMAPI Image_Performance_Query
        #   API.  The response is not in the expected format.
        #   The exception that occurred during processing of the
        #   response, its details and the response are included
        #   in the message.
        # SysAct: Processing of the function terminates.
        # UserResp: Determine the cause of the failure.  A code change
        #   may be needed in the function or in the z/VM SMAPI code.
        #   After correcting the code, reinvoke the function.
    '0413': [{'overallRC': 99, 'rc': 99, 'rs': 413},
            "ULT%s0413E Userid '%s' did not enter the expected " +
            "operating system state of '%s' in %i seconds.",
            ('GUT', 'USERID', 'DESIRED_STATE', 'MAX_WAIT')],
        # Explain: The managed system did not enter the operating
        #   system state that was shown in the message in the
        #   maximum number of seconds allowed for this to happen.
        #   The maximum number of seconds a combination of the,
        #   specified or defaulted, polling interval and maximum
        #   maximum number of polling attempts.
        # SysAct: Processing of the function terminates.
        # UserResp: Determine the cause of the failure and correct
        #   the cause.
    '0414': [{'overallRC': 99, 'rc': 99, 'rs': 414},
            "ULT%s0414E Userid '%s' did not enter the expected " +
            "virtual machine state of '%s' in %i seconds.",
            ('GUT', 'USERID', 'DESIRED_STATE', 'MAX_WAIT')],
        # Explain: The managed system did not enter the virtual
        #   machine log on/off state that was shown in the message
        #   in the maximum number of seconds allowed for this to happen.
        #   The maximum number of seconds a combination of the,
        #   specified or defaulted, polling interval and maximum
        #   maximum number of polling attempts.
        # SysAct: Processing of the function terminates.
        # UserResp: Determine the cause of the failure and correct
        #   the cause.
    '0415': [{'overallRC': 3, 'rc': 415},    # rs comes from failing rc
            "ULT%s0415E Command failed: '%s', rc: %i out: %s",
            ('GUT', 'CMD', 'RC', 'OUTPUT')],
        # Explain: The indicated command failed.  The return code
        #   and output from the command are shown.
        # SysAct: Function processing terminates.
        # UserResp: Use the information provided with the message
        #   to determine the cause of the failure and correct the
        #   problem.  Reinvoke the function after you correct the
        #   problem.
    '0416': [{'overallRC': 99, 'rc': 99, 'rs': 416},
            "ULT%s0416E Command returned a response " +
            "containing '%s' but did not have at least %i words " +
            "following it. cmd: '%s', out: '%s'",
            ('GUT', 'KEYWORD', 'NUM', 'CMD', 'OUTPUT')],
        # Explain: A command was invoked that returned a successful
        #   return code indication.  The response contained the
        #   expected string but did not contain the expected number
        #   of words that follow the string.
        # SysAct: Processing of the function terminates.
        # UserResp: Use the information provided in the message
        #   to determine the cause of the problem and correct it.
        #   Reinvoke the function after you correct the problem.
    '0417': [{'overallRC': 99, 'rc': 99, 'rs': 417},
            "ULT%s0417E Command did not return the expected response " +
            "containing '%s', cmd: '%s', out: '%s'",
            ('GUT', 'KEYWORD', 'CMD', 'OUTPUT')],
        # Explain: A command was invoked that returned a successful
        #   return code indication.  The response did not contain the
        #   expected string.
        # SysAct: Processing of the function terminates.
        # UserResp: Use the information provided in the message
        #   to determine the reason the identified string was not
        #   present in the response to identify the cause.
        #   Reinvoke the function after you correct the problem.
    '0418': [{'overallRC': 99, 'rc': 99, 'rs': 418},
            "ULT%s0418E Userid %s is not logged on to this system.",
            ('GUT', 'USERID')],
        # Explain: A CP message HCP0045E was returned, indicating the
        #   userid specified is not logged on to this z/VM system,
        #   thus it cannot be relocated.
        # SysAct:  Processing of the function ends with no action taken.
        # UserResp: Correct the function call to specify a correct userid and
        #   reinvoke the function.
    '0419': [{'overallRC': 99, 'rc': 99, 'rs': 419},
            "ULT%s0419E A relocation is not in progress for userid %s.",
            ('GUT', 'USERID')],
        # Explain: An attempt was made to query or cancel a relocation
        #   for a user, but the SMAPI command indicated that no
        #   relocation was in progress.
        # SysAct:  Processing of the function ends with no action taken.
        # UserResp: Reinvoke the function for a relocation that is in
        #   progress.
    '0420': [{'overallRC': 99, 'rc': 99, 'rs': 420},     # dict is not used.
            "ULT%s0420E An error occurred issuing a %s for userid %s. " +
            "Please look up message(s): %s in the CP Messages book for " +
            "more information.",
            ('GUT', 'CMD', 'USERID', 'ERROR_CODE')],
        # Explain: The VMRELOCATE command returns a list of messages
        #   containing all the problems encountered when trying to issue
        #   the command.
        # SysAct:   Processing of the function ends with no action taken.
        # UserResp: Look up the codes provided in the CP messages book,
        #   correct the problems and reinvoke the function.
    '0421': [{'overallRC': 99, 'rc': 421, 'rs': 0},
            "ULT%s0421E Exception received on an attempt to " +
            "execute a cmd: %s, exception: %s, " +
            "details: %s",
            ('GUT', 'CMD', 'EXCEPTION', 'EXCEPTION_DETAILS')],
        # Explain: The command indicated by the message failed.
        #   The error message contains exception name and details
        #   contained in the exception.
        # SysAct:   Processing of the function ends with no further
        #   action taken.
        # UserResp: Use the information in the message to determine
        #   the cause of the error and correct the problem.
        #   Reinvoke the function after you have corrected the problem.
    '0422': [{'overallRC': 99, 'rc': 422, 'rs': 0},
            "ULT%s0422W Exception received on an attempt to " +
            "execute a cmd: %s, exception: %s, " +
            "details: %s. Will attempt to continue processing.",
            ('GUT', 'CMD', 'EXCEPTION', 'EXCEPTION_DETAILS')],
        # Explain: While trying to execute a vmcp command, an error
        #   occurred. However the vmcp command was not central
        #   to processing the subfunction, so processing
        #   continues.
        # SysAct: Function processing continues.
        # UserResp: If there is reason to suspect the function did
        #   not execute completely, investigate the error.  Otherwise
        #   ignore this message.
    '0423': [{'overallRC': 4, 'rc': 4, 'rs': 423},    # dict is not used.
            "ULT%s0423W Unable to spool reader to all classes, " +
            "it is possible that there may be additional console " +
            "files available that are not listed in the response. " +
            "Command: %s, exception %s, details %s.  Will attempt " +
            "to continue processing.",
            ('GUT', 'CMD', 'EXCEPTION', 'EXCEPTION_DETAILS')],
        # Explain: The vmcp spool reader class * command was not
        #   successful.  This means the reader could not be changed
        #   to get files of all classes, and thus there could be
        #   files that are ignored.  The exception was of a different
        #   type than in message 407.
        # SysAct: Processing of the function continues.
        # UserResp: If missing files are suspected, investigate the
        #   cause of the failure in the CP messages book or vmcp
        #   help and reinvoke the function.
    '0424': [{'overallRC': 4, 'rc': 4, 'rs': 424},
            "ULT%s0424E Failed to transfer %s file to guest: %s, out: %s",
            ('GUT', 'FILE_LOCATION', 'USERID', 'OUTPUT')],
        # Explain: The vmcp transfer command failed for the specified
        #   reason.
        # SysAct: Processing of the function ends with no action
        #   taken.
        # UserResp: Look up the reason the vmcp transfer command failed,
        #   correct the problem and reinvoke the function.
    # 5000-6100: Reserved for SMCLI
    }



# Copyright 2017 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from zvmsdk import monitor
from zvmsdk.tests.unit import base

CPUMEM_SAMPLE1 = {
            'userid': 'USERID1',
            'guest_cpus': '1',
            'used_cpu_time': '6185838 uS',
            'elapsed_cpu_time': '35232895 uS',
            'min_cpu_count': '2',
            'max_cpu_limit': '10000',
            'samples_cpu_in_use': '0',
            'samples_cpu_delay': '0',
            'used_memory': '290232 KB',
            'max_memory': '2097152 KB',
            'min_memory': '0 KB',
            'shared_memory': '5222192 KB',
            }
CPUMEM_SAMPLE2 = {
            'userid': 'USERID2',
            'guest_cpus': '3',
            'used_cpu_time': '14293629 uS',
            'elapsed_cpu_time': '4868976371 uS',
            'min_cpu_count': '3',
            'max_cpu_limit': '10000',
            'samples_cpu_in_use': '0',
            'samples_cpu_delay': '0',
            'used_memory': '305020 KB',
            'max_memory': '2097152 KB',
            'min_memory': '0 KB',
            'shared_memory': '5222190 KB',
            }

MEM_KEYS = ['used_mem_kb', 'max_mem_kb', 'min_mem_kb', 'shared_mem_kb']

SMCLI_VSW_NIC_DATA = {'vswitches': [
            {'vswitch_name': 'TESTVSW1',
                'nics': [
                    {'nic_fr_rx_dsc': '0',
                     'nic_fr_rx_err': '0',
                     'nic_fr_tx_err': '4',
                     'userid': 'USERID1',
                     'nic_rx': '103024058',
                     'nic_fr_rx': '573952',
                     'nic_fr_tx': '548780',
                     'vdev': '0600',
                     'nic_fr_tx_dsc': '0',
                     'nic_tx': '102030890'},
                    {'nic_fr_rx_dsc': '0',
                     'nic_fr_rx_err': '0',
                     'nic_fr_tx_err': '4',
                     'userid': 'USERID2',
                     'nic_rx': '3111714',
                     'nic_fr_rx': '17493',
                     'nic_fr_tx': '16886',
                     'vdev': '0600',
                     'nic_fr_tx_dsc': '0',
                     'nic_tx': '3172646'}]},
            {'vswitch_name': 'TESTVSW2',
                'nics': [
                    {'nic_fr_rx_dsc': '0',
                     'nic_fr_rx_err': '0',
                     'nic_fr_tx_err': '0',
                     'userid': 'USERID1',
                     'nic_rx': '4684435',
                     'nic_fr_rx': '34958',
                     'nic_fr_tx': '16211',
                     'vdev': '1000',
                     'nic_fr_tx_dsc': '0',
                     'nic_tx': '3316601'},
                    {'nic_fr_rx_dsc': '0',
                     'nic_fr_rx_err': '0',
                     'nic_fr_tx_err': '0',
                     'userid': 'USERID2',
                     'nic_rx': '3577163',
                     'nic_fr_rx': '27211',
                     'nic_fr_tx': '12344',
                     'vdev': '1000',
                     'nic_fr_tx_dsc': '0',
                     'nic_tx': '2515045'}]}],
            'vswitch_count': 2}

INST_NICS_SAMPLE1 = [
            {'nic_fr_rx': 573952,
             'nic_fr_rx_dsc': 0,
             'nic_fr_rx_err': 0,
             'nic_fr_tx': 548780,
             'nic_fr_tx_dsc': 0,
             'nic_fr_tx_err': 4,
             'nic_rx': 103024058,
             'nic_tx': 102030890,
             'nic_vdev': '0600',
             'vswitch_name': 'TESTVSW1'},
            {'nic_fr_rx': 34958,
             'nic_fr_rx_dsc': 0,
             'nic_fr_rx_err': 0,
             'nic_fr_tx': 16211,
             'nic_fr_tx_dsc': 0,
             'nic_fr_tx_err': 0,
             'nic_rx': 4684435,
             'nic_tx': 3316601,
             'nic_vdev': '1000',
             'vswitch_name': 'TESTVSW2'}
        ]
INST_NICS_SAMPLE2 = [
            {'nic_fr_rx': 17493,
             'nic_fr_rx_dsc': 0,
             'nic_fr_rx_err': 0,
             'nic_fr_tx': 16886,
             'nic_fr_tx_dsc': 0,
             'nic_fr_tx_err': 4,
             'nic_rx': 3111714,
             'nic_tx': 3172646,
             'nic_vdev': '0600',
             'vswitch_name': 'TESTVSW1'},
            {'nic_fr_rx': 27211,
             'nic_fr_rx_dsc': 0,
             'nic_fr_rx_err': 0,
             'nic_fr_tx': 12344,
             'nic_fr_tx_dsc': 0,
             'nic_fr_tx_err': 0,
             'nic_rx': 3577163,
             'nic_tx': 2515045,
             'nic_vdev': '1000',
             'vswitch_name': 'TESTVSW2'}
        ]


class SDKMonitorTestCase(base.SDKTestCase):
    def setUp(self):
        self._monitor = monitor.ZVMMonitor()

    @mock.patch("zvmsdk.monitor.MeteringCache.get")
    @mock.patch("zvmsdk.smutclient.SMUTClient.get_power_state")
    @mock.patch("zvmsdk.monitor.ZVMMonitor._cache_enabled")
    def test_private_get_inspect_data_cache_hit_single(self, cache_enabled,
                                                       get_ps, cache_get):
        cache_get.return_value = CPUMEM_SAMPLE1
        rdata = self._monitor._get_inspect_data('cpumem', ['USERID1'])
        self.assertEqual(list(rdata.keys()), ['USERID1'])
        self.assertEqual(sorted(list(rdata['USERID1'].keys())),
                         sorted(CPUMEM_SAMPLE1.keys()))
        self.assertEqual(rdata['USERID1']['guest_cpus'], '1')
        self.assertEqual(rdata['USERID1']['used_cpu_time'], '6185838 uS')
        self.assertEqual(rdata['USERID1']['used_memory'], '290232 KB')
        self.assertEqual(rdata['USERID1']['shared_memory'], '5222192 KB')
        get_ps.assert_not_called()
        cache_enabled.assert_not_called()

    @mock.patch("zvmsdk.monitor.MeteringCache.get")
    @mock.patch("zvmsdk.smutclient.SMUTClient.get_power_state")
    @mock.patch("zvmsdk.monitor.ZVMMonitor._cache_enabled")
    def test_private_get_inspect_data_cache_hit_multi(self, cache_enabled,
                                                       get_ps, cache_get):
        cache_get.side_effect = [CPUMEM_SAMPLE1,
            CPUMEM_SAMPLE2]
        rdata = self._monitor._get_inspect_data('cpumem',
                                                ['USERID1', 'USERID2'])
        self.assertEqual(sorted(rdata.keys()), ['USERID1', 'USERID2'])
        self.assertEqual(sorted(rdata['USERID1'].keys()),
                         sorted(CPUMEM_SAMPLE1.keys()))
        self.assertEqual(rdata['USERID1']['guest_cpus'], '1')
        self.assertEqual(rdata['USERID1']['used_cpu_time'], '6185838 uS')
        self.assertEqual(rdata['USERID1']['used_memory'], '290232 KB')
        self.assertEqual(rdata['USERID1']['shared_memory'], '5222192 KB')
        self.assertEqual(rdata['USERID2']['guest_cpus'], '3')
        self.assertEqual(rdata['USERID2']['used_cpu_time'], '14293629 uS')
        self.assertEqual(rdata['USERID2']['used_memory'], '305020 KB')
        self.assertEqual(rdata['USERID2']['shared_memory'], '5222190 KB')
        get_ps.assert_not_called()
        cache_enabled.assert_not_called()

    @mock.patch("zvmsdk.monitor.MeteringCache.get")
    @mock.patch("zvmsdk.smutclient.SMUTClient.get_power_state")
    @mock.patch("zvmsdk.monitor.ZVMMonitor._update_cpumem_data")
    def test_private_get_inspect_data_cache_miss_single(self,
                                                        update_cpumem_data,
                                                        get_ps, cache_get):
        cache_get.return_value = None
        get_ps.return_value = 'on'
        update_cpumem_data.return_value = {
            'USERID1': CPUMEM_SAMPLE1,
            'USERID2': CPUMEM_SAMPLE2
            }
        rdata = self._monitor._get_inspect_data('cpumem', ['userid1'])
        get_ps.assert_called_once_with('userid1')
        update_cpumem_data.assert_called_once_with(['userid1'])
        self.assertEqual(sorted(rdata.keys()), sorted(['USERID1', 'USERID2']))
        self.assertEqual(sorted(rdata['USERID1'].keys()),
                         sorted(CPUMEM_SAMPLE1.keys()))
        self.assertEqual(rdata['USERID1']['guest_cpus'], '1')
        self.assertEqual(rdata['USERID1']['used_cpu_time'], '6185838 uS')
        self.assertEqual(rdata['USERID1']['used_memory'], '290232 KB')

    @mock.patch("zvmsdk.monitor.MeteringCache.get")
    @mock.patch("zvmsdk.smutclient.SMUTClient.get_power_state")
    @mock.patch("zvmsdk.monitor.ZVMMonitor._update_cpumem_data")
    def test_private_get_inspect_data_cache_miss_multi(self,
                                                        update_cpumem_data,
                                                        get_ps, cache_get):
        cache_get.side_effect = [{
            'userid': 'USERID1',
            'guest_cpus': '1',
            'used_cpu_time': '7185838 uS',
            'elapsed_cpu_time': '35232895 uS',
            'min_cpu_count': '2',
            'max_cpu_limit': '10000',
            'samples_cpu_in_use': '0',
            'samples_cpu_delay': '0',
            'used_memory': '390232 KB',
            'max_memory': '2097152 KB',
            'min_memory': '0 KB',
            'shared_memory': '4222192 KB',
            }, None]
        get_ps.return_value = 'on'
        update_cpumem_data.return_value = {
            'USERID1': CPUMEM_SAMPLE1,
            'USERID2': CPUMEM_SAMPLE2
            }
        rdata = self._monitor._get_inspect_data('cpumem',
                                                ['userid1', 'userid2'])
        get_ps.assert_called_once_with('userid2')
        update_cpumem_data.assert_called_once_with(['userid1', 'userid2'])
        self.assertEqual(sorted(rdata.keys()), sorted(['USERID1', 'USERID2']))
        self.assertEqual(sorted(rdata['USERID1'].keys()),
                         sorted(CPUMEM_SAMPLE1.keys()))
        self.assertEqual(rdata['USERID1']['guest_cpus'], '1')
        self.assertEqual(rdata['USERID1']['used_cpu_time'], '6185838 uS')
        self.assertEqual(rdata['USERID1']['used_memory'], '290232 KB')
        self.assertEqual(rdata['USERID1']['shared_memory'], '5222192 KB')

    @mock.patch("zvmsdk.monitor.MeteringCache.get")
    @mock.patch("zvmsdk.smutclient.SMUTClient.get_power_state")
    @mock.patch("zvmsdk.monitor.ZVMMonitor._update_cpumem_data")
    def test_private_get_inspect_data_guest_off(self,
                                                update_cpumem_data,
                                                get_ps, cache_get):
        cache_get.return_value = None
        get_ps.return_value = 'off'
        rdata = self._monitor._get_inspect_data('cpumem',
                                                ['userid1'])
        get_ps.assert_called_once_with('userid1')
        update_cpumem_data.assert_not_called()
        self.assertEqual(rdata, {})

    @mock.patch("zvmsdk.monitor.MeteringCache.get")
    @mock.patch("zvmsdk.smutclient.SMUTClient.get_power_state")
    @mock.patch("zvmsdk.monitor.ZVMMonitor._update_nic_data")
    def test_private_get_inspect_data_vnics(self,
                                            update_nic_data,
                                            get_ps, cache_get):
        cache_get.return_value = None
        get_ps.return_value = 'on'
        update_nic_data.return_value = {'USERID1': INST_NICS_SAMPLE1,
                                        'USERID2': INST_NICS_SAMPLE2
                                        }
        rdata = self._monitor._get_inspect_data('vnics',
                                                ['USERID1'])
        get_ps.assert_called_once_with('USERID1')
        update_nic_data.assert_called_once_with()
        self.assertEqual(rdata, {'USERID1': INST_NICS_SAMPLE1,
                                 'USERID2': INST_NICS_SAMPLE2
                                 })

    @mock.patch("zvmsdk.smutclient.SMUTClient.system_image_performance_query")
    @mock.patch("zvmsdk.monitor.ZVMMonitor._cache_enabled")
    @mock.patch("zvmsdk.smutclient.SMUTClient.get_vm_list")
    @mock.patch("zvmsdk.smutclient.SMUTClient.namelist_query")
    def test_private_update_cpumem_data_cache_enabled(self, namelist_query,
                                               get_vm_list,
                                               cache_enabled,
                                               image_performance_query):
        cache_enabled.return_value = True
        namelist_query.return_value = ['USERID1', 'USERID2']
        get_vm_list.return_value = ['USERID1', 'USERID2']
        image_performance_query.return_value = {
            'USERID1': CPUMEM_SAMPLE1,
            'USERID2': CPUMEM_SAMPLE2
            }
        rdata = self._monitor._update_cpumem_data(['userid1'])
        image_performance_query.assert_called_once_with('TSTNLIST')
        namelist_query.assert_called_once_with('TSTNLIST')
        get_vm_list.assert_called_once_with()
        self.assertEqual(sorted(rdata.keys()), sorted(['USERID1', 'USERID2']))
        self.assertEqual(rdata['USERID1']['guest_cpus'], '1')
        self.assertEqual(rdata['USERID1']['used_cpu_time'], '6185838 uS')
        self.assertEqual(rdata['USERID1']['used_memory'], '290232 KB')
        self.assertEqual(
        self._monitor._cache._cache['cpumem']['data']['USERID2']['guest_cpus'],
        '3')

    @mock.patch("zvmsdk.smutclient.SMUTClient.system_image_performance_query")
    @mock.patch("zvmsdk.smutclient.SMUTClient.namelist_add")
    @mock.patch("zvmsdk.smutclient.SMUTClient.get_vm_list")
    @mock.patch("zvmsdk.smutclient.SMUTClient.namelist_query")
    @mock.patch("zvmsdk.monitor.ZVMMonitor._cache_enabled")
    def test_private_update_cpumem_data_cache_not_in_namelist(self,
                                cache_enabled, namelist_query, get_vm_list,
                                namelist_add, image_performance_query):
        cache_enabled.return_value = True
        namelist_query.return_value = ['USERID1']
        get_vm_list.return_value = ['USERID1', 'USERID2']
        image_performance_query.return_value = {
            'USERID1': CPUMEM_SAMPLE1,
            'USERID2': CPUMEM_SAMPLE2,
            }
        rdata = self._monitor._update_cpumem_data(['USERID1', 'USERID2'])
        image_performance_query.assert_called_once_with('TSTNLIST')
        namelist_query.assert_called_once_with('TSTNLIST')
        get_vm_list.assert_called_once_with()
        namelist_add.assert_called_once_with('TSTNLIST', 'USERID2')
        self.assertEqual(sorted(rdata.keys()), sorted(['USERID1', 'USERID2']))
        self.assertEqual(rdata['USERID1']['guest_cpus'], '1')
        self.assertEqual(rdata['USERID1']['used_cpu_time'], '6185838 uS')
        self.assertEqual(rdata['USERID1']['used_memory'], '290232 KB')
        self.assertEqual(
        self._monitor._cache._cache['cpumem']['data']['USERID2']['guest_cpus'],
        '3')

    @mock.patch("zvmsdk.smutclient.SMUTClient.get_vm_list")
    @mock.patch("zvmsdk.monitor.ZVMMonitor._cache_enabled")
    @mock.patch("zvmsdk.smutclient.SMUTClient.system_image_performance_query")
    @mock.patch("zvmsdk.smutclient.SMUTClient.namelist_query")
    def test_private_update_cpumem_data_cache_disabled(self, namelist_query,
                                                image_perform_query,
                                                cache_enabled, get_vm_list):
        namelist_query.return_value = ['USERID1', 'USERID2']
        cache_enabled.return_value = False
        image_perform_query.return_value = {
            'USERID1': CPUMEM_SAMPLE1
            }
        get_vm_list.return_value = ['USERID1', 'USERID2']
        rdata = self._monitor._update_cpumem_data(['userid1'])
        namelist_query.assert_called_once_with('TSTNLIST')
        get_vm_list.assert_called_once_with()
        image_perform_query.assert_called_once_with('TSTNLIST')
        self.assertEqual(list(rdata.keys()), ['USERID1'])
        self.assertEqual(sorted(rdata['USERID1'].keys()),
                         sorted(CPUMEM_SAMPLE1.keys()))
        self.assertEqual(rdata['USERID1']['guest_cpus'], '1')
        self.assertEqual(rdata['USERID1']['used_cpu_time'], '6185838 uS')
        self.assertEqual(rdata['USERID1']['used_memory'], '290232 KB')
        self.assertEqual(
            list(self._monitor._cache._cache['cpumem']['data'].keys()), [])

    @mock.patch("zvmsdk.monitor.ZVMMonitor._get_inspect_data")
    def test_inspect_stats_single(self, _get_inspect_data):
        _get_inspect_data.return_value = {
            'USERID1': CPUMEM_SAMPLE1,
            'USERID2': CPUMEM_SAMPLE2
            }
        rdata = self._monitor.inspect_stats(['USERID1'])
        _get_inspect_data.assert_called_once_with('cpumem', ['USERID1'])
        self.assertEqual(sorted(rdata.keys()), sorted(['USERID1']))
        self.assertEqual(rdata['USERID1']['guest_cpus'], 1)
        self.assertEqual(rdata['USERID1']['used_cpu_time_us'], 6185838)
        self.assertEqual(rdata['USERID1']['elapsed_cpu_time_us'], 35232895)
        self.assertEqual(rdata['USERID1']['min_cpu_count'], 2)
        self.assertEqual(rdata['USERID1']['max_cpu_limit'], 10000)
        self.assertEqual(rdata['USERID1']['used_mem_kb'], 290232)
        self.assertEqual(rdata['USERID1']['max_mem_kb'], 2097152)
        self.assertEqual(rdata['USERID1']['min_mem_kb'], 0)
        self.assertEqual(rdata['USERID1']['shared_mem_kb'], 5222192)

    @mock.patch("zvmsdk.monitor.ZVMMonitor._get_inspect_data")
    def test_inspect_stats_multi(self, _get_inspect_data):
        _get_inspect_data.return_value = {
            'USERID1': CPUMEM_SAMPLE1,
            'USERID2': CPUMEM_SAMPLE2
            }
        rdata = self._monitor.inspect_stats(['USERID1', 'USERID2'])
        _get_inspect_data.assert_called_once_with('cpumem',
                                                  ['USERID1', 'USERID2'])
        self.assertEqual(sorted(rdata.keys()), sorted(['USERID1', 'USERID2']))
        self.assertEqual(rdata['USERID1']['guest_cpus'], 1)
        self.assertEqual(rdata['USERID1']['used_cpu_time_us'], 6185838)
        self.assertEqual(rdata['USERID1']['elapsed_cpu_time_us'], 35232895)
        self.assertEqual(rdata['USERID1']['min_cpu_count'], 2)
        self.assertEqual(rdata['USERID1']['max_cpu_limit'], 10000)
        self.assertEqual(rdata['USERID2']['guest_cpus'], 3)
        self.assertEqual(rdata['USERID2']['used_cpu_time_us'], 14293629)
        self.assertEqual(rdata['USERID2']['elapsed_cpu_time_us'], 4868976371)
        self.assertEqual(rdata['USERID2']['min_cpu_count'], 3)
        self.assertEqual(rdata['USERID2']['max_cpu_limit'], 10000)
        self.assertEqual(rdata['USERID1']['used_mem_kb'], 290232)
        self.assertEqual(rdata['USERID1']['max_mem_kb'], 2097152)
        self.assertEqual(rdata['USERID1']['min_mem_kb'], 0)
        self.assertEqual(rdata['USERID1']['shared_mem_kb'], 5222192)
        self.assertEqual(rdata['USERID2']['used_mem_kb'], 305020)
        self.assertEqual(rdata['USERID2']['max_mem_kb'], 2097152)
        self.assertEqual(rdata['USERID2']['min_mem_kb'], 0)
        self.assertEqual(rdata['USERID2']['shared_mem_kb'], 5222190)

    @mock.patch("zvmsdk.monitor.ZVMMonitor._get_inspect_data")
    def test_inspect_stats_single_off_or_not_exist(self, _get_inspect_data):
        _get_inspect_data.return_value = {
            'USERID2': CPUMEM_SAMPLE2
            }
        rdata = self._monitor.inspect_stats(['userid1'])
        _get_inspect_data.assert_called_once_with('cpumem', ['userid1'])
        self.assertEqual(rdata, {})

    @mock.patch("zvmsdk.monitor.ZVMMonitor._get_inspect_data")
    def test_inspect_stats_multi_off_or_not_exist(self, _get_inspect_data):
        _get_inspect_data.return_value = {
            'USERID1': CPUMEM_SAMPLE1
            }
        rdata = self._monitor.inspect_stats(['USERID1', 'USERID2'])
        _get_inspect_data.assert_called_once_with('cpumem',
                                                  ['USERID1', 'USERID2'])
        self.assertEqual(sorted(rdata.keys()), sorted(['USERID1']))
        self.assertEqual(rdata['USERID1']['guest_cpus'], 1)
        self.assertEqual(rdata['USERID1']['used_cpu_time_us'], 6185838)
        self.assertEqual(rdata['USERID1']['elapsed_cpu_time_us'], 35232895)
        self.assertEqual(rdata['USERID1']['min_cpu_count'], 2)
        self.assertEqual(rdata['USERID1']['max_cpu_limit'], 10000)
        self.assertEqual(rdata['USERID1']['used_mem_kb'], 290232)
        self.assertEqual(rdata['USERID1']['max_mem_kb'], 2097152)
        self.assertEqual(rdata['USERID1']['min_mem_kb'], 0)
        self.assertEqual(rdata['USERID1']['shared_mem_kb'], 5222192)

    @mock.patch("zvmsdk.smutclient.SMUTClient"
                ".virtual_network_vswitch_query_byte_stats")
    @mock.patch("zvmsdk.monitor.ZVMMonitor._cache_enabled")
    def test_private_update_nic_data(self, cache_enabled, smcli_iuo_query):
        smcli_iuo_query.return_value = SMCLI_VSW_NIC_DATA
        cache_enabled.return_value = True
        nics_dict = self._monitor._update_nic_data()
        self.assertEqual(sorted(["USERID1", "USERID2"]),
                         sorted(nics_dict.keys()))
        self.assertEqual(nics_dict['USERID1'], INST_NICS_SAMPLE1)
        self.assertEqual(nics_dict['USERID2'], INST_NICS_SAMPLE2)
        self.assertEqual(self._monitor._cache.get('vnics', 'USERID1'),
                         INST_NICS_SAMPLE1)
        self.assertEqual(self._monitor._cache.get('vnics', 'USERID2'),
                         INST_NICS_SAMPLE2)

    @mock.patch("zvmsdk.smutclient.SMUTClient"
                ".virtual_network_vswitch_query_byte_stats")
    @mock.patch("zvmsdk.monitor.ZVMMonitor._cache_enabled")
    def test_private_update_nic_data_cache_disabled(self, cache_enabled,
                                                    smcli_iuo_query):
        smcli_iuo_query.return_value = SMCLI_VSW_NIC_DATA
        cache_enabled.return_value = False
        nics_dict = self._monitor._update_nic_data()
        self.assertEqual(sorted(["USERID1", "USERID2"]),
                         sorted(nics_dict.keys()))
        self.assertEqual(self._monitor._cache.get('vnics', 'USERID1'),
                         None)
        self.assertEqual(self._monitor._cache.get('vnics', 'USERID2'),
                         None)



# Copyright 2017,2018 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import six

import webob
from webob.dec import wsgify

from zvmsdk import log


LOG = log.LOG
SDKWSGI_MODID = 120


def extract_json(body):
    try:
        LOG.debug('Decoding body: %s', body)
        # This function actually is received from upper layer through
        # socket, so it's bytes in py3
        if isinstance(body, bytes):
            body = bytes.decode(body)

        data = json.loads(body)
    except ValueError as exc:
        msg = ('Malformed JSON: %(error)s') % {'error': exc}
        LOG.debug(msg)
        raise webob.exc.HTTPBadRequest(msg,
            json_formatter=json_error_formatter)
    return data


def json_error_formatter(body, status, title, environ):
    """A json_formatter for webob exceptions."""
    body = webob.exc.strip_tags(body)
    status_code = int(status.split(None, 1)[0])
    error_dict = {
        'status': status_code,
        'title': title,
        'detail': body
    }
    return {'errors': [error_dict]}


def wsgi_path_item(environ, name):
    """Extract the value of a named field in a URL.

    Return None if the name is not present or there are no path items.
    """
    try:
        return environ['wsgiorg.routing_args'][1][name]
    except (KeyError, IndexError):
        return None


TRUE_STRINGS = ('1', 't', 'true', 'on', 'y', 'yes')
FALSE_STRINGS = ('0', 'f', 'false', 'off', 'n', 'no')


def bool_from_string(subject, strict=False, default=False):
    if isinstance(subject, bool):
        return subject
    if not isinstance(subject, six.string_types):
        subject = six.text_type(subject)

    lowered = subject.strip().lower()

    if lowered in TRUE_STRINGS:
        return True
    elif lowered in FALSE_STRINGS:
        return False
    elif strict:
        acceptable = ', '.join(
            "'%s'" % s for s in sorted(TRUE_STRINGS + FALSE_STRINGS))
        msg = ("Unrecognized value '%(val)s', acceptable values are:"
               " %(acceptable)s") % {'val': subject,
                                     'acceptable': acceptable}
        raise ValueError(msg)
    else:
        return default


def get_request_uri(environ):
    name = environ.get('SCRIPT_NAME', '')
    info = environ.get('PATH_INFO', '')
    req_uri = name + info
    if environ.get('QUERY_STRING'):
        req_uri += '?' + environ['QUERY_STRING']
    return req_uri


def get_http_code_from_sdk_return(msg, additional_handler=None, default=200):
    LOG.debug("Get msg to handle: %s", msg)

    if 'overallRC' in msg:
        ret = msg['overallRC']

        if ret != 0:
            # same definition to sdk layer
            if ret in [400, 404, 409, 501, 503]:
                return ret

            # 100 mean validation error in sdk layer and
            # lead to a 400 badrequest
            if ret in [100]:
                return 400

            # Add a special handle for smut return
            if additional_handler:
                ret = additional_handler(msg)
                if ret:
                    return ret

            # ok, we reach here because can't handle it
            LOG.info("The msg <%s> lead to return internal error", msg)
            return 500
        else:
            # return default code
            return default


def handle_not_found(msg):

    if 'overallRC' in msg and 'rc' in msg and 'rs' in msg:
        # overall rc: 8, rc: 212, rs: 40 means vswitch not exist
        if (msg['overallRC'] == 8 and msg['rc'] == 212 and
            msg['rs'] == 40):
            LOG.debug('vswitch does not exist, change ret to 404')
            return 404

        # overall rc: 4, rc: 5, rs: 402 means vswitch not exist
        if (msg['overallRC'] == 4 and msg['rc'] == 5 and
            msg['rs'] == 402):
            LOG.debug('disk pool not exist, change ret to 404')
            return 404

        # overall rc: 300, rc: 300, rs: 20 means image not exist
        if (msg['overallRC'] == 300 and msg['rc'] == 300 and
            msg['rs'] == 20):
            LOG.debug('image not exist, change ret to 404')
            return 404

        # overall rc: 8, rc: 400, rs: 4 means guest not exist
        if (msg['overallRC'] == 8 and msg['rc'] == 400 and
            msg['rs'] == 4):
            LOG.debug('guest not exist, change ret to 404')
            return 404

        # overall rc: 8, rc: 200, rs: 4 means guest not exist
        if (msg['overallRC'] == 8 and msg['rc'] == 200 and
            msg['rs'] == 4):
            LOG.debug('guest not exist, change ret to 404')
            return 404

        # overall rc: 300, rc:300, rs: 3, error message contains
        # "not linked; not in CP directory" means target vdev not exist
        if (msg['overallRC'] == 300 and msg['rc'] == 300 and
            msg['rs'] == 3 and 'not linked; not in CP directory' in
            msg['errmsg']):
            LOG.debug('deploy target vdev not exist,'
                      ' change ret to 404')
            return 404
    return 0


def handle_already_exists(msg):
    if 'overallRC' in msg and 'rc' in msg and 'rs' in msg:
        # overall rc: 8, rc: 212, rs: 36 means vswitch already exist
        if (msg['overallRC'] == 8 and msg['rc'] == 212 and
            msg['rs'] == 36):
            LOG.debug('vswitch already exist, change ret to 409')
            return 409

        # overall rc: 300, rc: 300, rc: 13 means image already exist
        if (msg['overallRC'] == 300 and msg['rc'] == 300 and
            msg['rs'] == 13):
            LOG.debug('image already exist, change ret to 409')
            return 409

        # overall rc: 8, rc: 400, rs: 8 means guest already exist
        if (msg['overallRC'] == 8 and msg['rc'] == 400 and
            msg['rs'] == 8):
            LOG.debug('guest already exist, change ret to 409')
            return 409

    # not handle it well, go to default
    return 0


class SdkWsgify(wsgify):

    def call_func(self, req, *args, **kwargs):
        """Add json_error_formatter to any webob HTTPExceptions."""
        try:
            return super(SdkWsgify, self).call_func(req, *args, **kwargs)
        except webob.exc.HTTPException as exc:
            msg = ('encounter %(error)s error') % {'error': exc}
            LOG.debug(msg)
            exc.json_formatter = json_error_formatter
            code = exc.status_int
            explanation = six.text_type(exc)

            fault_data = {
                'overallRC': 400,
                'rc': 400,
                'rs': code,
                'modID': SDKWSGI_MODID,
                'output': '',
                'errmsg': explanation}
            exc.text = six.text_type(json.dumps(fault_data))
            raise exc
